#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union

from argilla.client.feedback.schemas.fields import TextField
from argilla.client.feedback.schemas.metadata import (
    FloatMetadataProperty,
    IntegerMetadataProperty,
    TermsMetadataProperty,
)
from argilla.client.feedback.schemas.questions import (
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    TextQuestion,
)
from argilla.client.feedback.schemas.remote.fields import RemoteTextField
from argilla.client.feedback.schemas.remote.metadata import (
    RemoteFloatMetadataProperty,
    RemoteIntegerMetadataProperty,
    RemoteTermsMetadataProperty,
)
from argilla.client.feedback.schemas.remote.questions import (
    RemoteLabelQuestion,
    RemoteMultiLabelQuestion,
    RemoteRankingQuestion,
    RemoteRatingQuestion,
    RemoteTextQuestion,
)
from argilla.client.feedback.schemas.remote.vector_settings import RemoteVectorSettings
from argilla.client.feedback.schemas.vector_settings import VectorSettings

AllowedFieldTypes = TextField
AllowedRemoteFieldTypes = RemoteTextField
AllowedQuestionTypes = Union[TextQuestion, RatingQuestion, LabelQuestion, MultiLabelQuestion, RankingQuestion]
AllowedRemoteQuestionTypes = Union[
    RemoteTextQuestion, RemoteRatingQuestion, RemoteLabelQuestion, RemoteMultiLabelQuestion, RemoteRankingQuestion
]
AllowedMetadataPropertyTypes = Union[TermsMetadataProperty, FloatMetadataProperty, IntegerMetadataProperty]
AllowedRemoteMetadataPropertyTypes = Union[
    RemoteTermsMetadataProperty, RemoteIntegerMetadataProperty, RemoteFloatMetadataProperty
]
AllowedVectorSettingsTypes = VectorSettings
AllowedRemoteVectorSettingsTypes = RemoteVectorSettings



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import textwrap
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from argilla.client.feedback.schemas.records import FeedbackRecord
from argilla.client.feedback.training.schemas.base import TrainingTaskForTextClassification, TrainingTaskTypes
from argilla.client.models import Framework, TextClassificationRecord
from argilla.training.base import ArgillaTrainer as ArgillaTrainerV1
from argilla.training.base import ArgillaTrainerSkeleton as ArgillaTrainerSkeletonV1

if TYPE_CHECKING:
    import spacy
    from transformers import PreTrainedModel, PreTrainedTokenizer

    from argilla.client.feedback.dataset.local.dataset import FeedbackDataset
    from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackDataset
    from argilla.client.feedback.integrations.huggingface.model_card import ArgillaModelCard
    from argilla.client.feedback.schemas.enums import ResponseStatusFilter
    from argilla.client.feedback.schemas.records import SortBy


class ArgillaTrainer(ArgillaTrainerV1):
    def __init__(
        self,
        dataset: ["FeedbackDataset", "RemoteFeedbackDataset"],
        task: TrainingTaskTypes,
        framework: Framework,
        lang: Optional["spacy.Language"] = None,
        model: Optional[Union[str, "PreTrainedModel"]] = None,
        tokenizer: Optional["PreTrainedTokenizer"] = None,
        train_size: Optional[float] = None,
        seed: Optional[int] = None,
        gpu_id: Optional[int] = -1,
        filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
        sort_by: Optional[List["SortBy"]] = None,
        max_records: Optional[int] = None,
        framework_kwargs: Optional[dict] = {},
    ) -> None:
        """
        Initialize an Argilla Trainer.

        Args:
            dataset: the dataset to be used for training.
            task: the training data to be used for training.
            framework: the framework to use for training. Currently, "transformers", "setfit", "spacy", "peft",
                "openai", "span_marker", "trl" and "sentence-transformers" are supported.
            lang: the spaCy language model to use for training, just required when `framework="spacy"`.
                Defaults to None, but it will be set to `spacy.blank("en")` if not specified.
            model: name or path to the baseline model to be used, or a transformers PreTrainedModel instance
                if the framework is "transformers", "peft" or "trl". If not specified will set to a good default
                per framework, if applicable. Defaults to None.
            tokenizer: a transformers PreTrainedTokenizer instance to tokenize the text. Only used with the
                "transformers", "peft" or "trl" frameworks.
            train_size: the size of the training set. If not specified, the entire dataset will be used for training,
                which may be an issue if `framework="spacy"` as it requires a validation set. Defaults to None.
            seed: the random seed to ensure reproducibility. Defaults to None.
            gpu_id: the GPU ID to use when training a SpaCy model. Defaults to -1, which means that the CPU
                will be used by default. GPU IDs start in 0, which stands for the default GPU in the system,
                if available.
            filter_by: A dict with key the field to filter by, and values the filters to apply. Currently only
                defined for `response_status` filters. Can be one of: draft, pending, submitted, and discarded.
                Defaults to `None` (no filter is applied).
            sort_by: A list of `SortBy` objects to sort your dataset by.
                Defaults to `None` (no filter is applied).
            max_records: the maximum number of records to use for training. Defaults to None.
            framework_kwargs: arguments for the framework's trainer. A special key (model_card_kwargs) is reserved
                for the arguments that can be passed to the model card.
            **load_kwargs: arguments for the rg.load() function.
        """
        if filter_by:
            dataset = dataset.filter_by(**filter_by)
        if sort_by:
            dataset = dataset.sort_by(sort_by)
        if max_records:
            dataset = dataset.pull(max_records=max_records)

        self._dataset = dataset
        self._train_size = train_size
        self._task = task
        self._seed = seed  # split is used for train-test-split and should therefore be fixed
        self._model = model
        self._tokenizer = tokenizer

        self._prepared_data = self._dataset.prepare_for_training(
            framework=framework,
            task=task,
            train_size=train_size,
            seed=seed,
            lang=lang,
        )

        if isinstance(framework, str):
            framework = Framework(framework)

        if tokenizer and framework not in (Framework.TRANSFORMERS, Framework.PEFT, Framework.TRL):
            warnings.warn(
                f"Passing a tokenizer is not supported for the {framework} framework.", UserWarning, stacklevel=2
            )

        # Save the model_card arguments if given by the user
        if model_card_kwargs := framework_kwargs.pop("model_card_kwargs", None):
            self.model_card_kwargs = model_card_kwargs
        else:
            self.model_card_kwargs = {}

        if framework is Framework.SETFIT:
            if not isinstance(task, TrainingTaskForTextClassification):
                raise NotImplementedError(f"{Framework.SETFIT} only supports `TextClassification` tasks.")
            from argilla.client.feedback.training.frameworks.setfit import ArgillaSetFitTrainer

            self._trainer = ArgillaSetFitTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.TRANSFORMERS:
            from argilla.client.feedback.training.frameworks.transformers import ArgillaTransformersTrainer

            self._trainer = ArgillaTransformersTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                tokenizer=self._tokenizer,
            )
        elif framework is Framework.PEFT:
            from argilla.client.feedback.training.frameworks.peft import ArgillaPeftTrainer

            self._trainer = ArgillaPeftTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                tokenizer=self._tokenizer,
            )
        elif framework is Framework.SPACY:
            from argilla.client.feedback.training.frameworks.spacy import ArgillaSpaCyTrainer

            self._trainer = ArgillaSpaCyTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                gpu_id=gpu_id,
                framework_kwargs=framework_kwargs,  # freeze_tok2vec
            )
        elif framework is Framework.SPACY_TRANSFORMERS:
            from argilla.client.feedback.training.frameworks.spacy import ArgillaSpaCyTransformersTrainer

            self._trainer = ArgillaSpaCyTransformersTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                gpu_id=gpu_id,
                framework_kwargs=framework_kwargs,  # update_transformer
            )
        elif framework is Framework.OPENAI:
            from argilla.client.feedback.training.frameworks.openai import ArgillaOpenAITrainer

            self._trainer = ArgillaOpenAITrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.SPAN_MARKER:
            from argilla.client.feedback.training.frameworks.span_marker import ArgillaSpanMarkerTrainer

            self._trainer = ArgillaSpanMarkerTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.TRL:
            from argilla.client.feedback.training.frameworks.trl import ArgillaTRLTrainer

            self._trainer = ArgillaTRLTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                tokenizer=self._tokenizer,
            )
        elif framework is Framework.SENTENCE_TRANSFORMERS:
            from argilla.client.feedback.training.frameworks.sentence_transformers import (
                ArgillaSentenceTransformersTrainer,
            )

            self._trainer = ArgillaSentenceTransformersTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                **framework_kwargs,  # cross_encoder
            )
        else:
            raise NotImplementedError(f"{framework} is not a valid framework.")

        self._logger.info(self)
        self._track_trainer_usage(framework=framework, task=self._task.__class__.__name__)

    @property
    def task(self) -> TrainingTaskTypes:
        """The task to be trained."""
        return self._task

    @property
    def trainer(
        self,
    ):
        return self._trainer

    def __repr__(self) -> str:
        """
        `trainer.__repr__()` prints out the trainer's parameters and a summary of how to use the trainer

        Returns:
          The trainer object.
        """
        return textwrap.dedent(
            f"""\
            ArgillaBaseTrainer info:
            _________________________________________________________________
            These baseline params are fixed:
                dataset: {self._dataset}
                task: {self._task}
                train_size: {self._train_size}
                seed: {self._seed}


            {self._trainer.__class__} info:
            _________________________________________________________________
            The parameters are configurable via `trainer.update_config()`:
                {self._trainer}

            Using the trainer:
            _________________________________________________________________
            `trainer.train(output_dir)` to train to start training. `output_dir` is the directory to save the model automatically.
            `trainer.predict(text, as_argilla_records=True)` to make predictions.
            `trainer.save(output_dir)` to save the model manually.
            """
        )

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        `predict` takes a string or list of strings and returns a list of dictionaries, each dictionary
        containing the text, the predicted label, and the confidence score.

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the output will be a list of Argilla records instead of dictionaries. Defaults to True.

        Returns:
          A list of predictions or Argilla records.
        """
        return self._trainer.predict(text=text, as_argilla_records=False, **kwargs)

    def save(self, output_dir: str, generate_card: bool = True) -> None:
        """
        Saves the model to the specified path and optionally generates a `ModelCard` at the same `output_dir`.

        Args:
            output_dir: The path to the directory where the model will be saved.
            generate_card: Whether to generate a model card of the `ArgillaTrainer` for the HuggingFace Hub. Defaults
                to `True`.
        """
        super().save(output_dir)

        if generate_card:
            self.generate_model_card(output_dir)

    def generate_model_card(self, output_dir: Optional[str] = None) -> "ArgillaModelCard":
        """Generate and return a model card based on the model card data.

        Args:
            output_dir: If given, folder where the model card will be written.

        Returns:
            model_card: The model card.
        """
        from argilla.client.feedback.integrations.huggingface.model_card import ArgillaModelCard

        if not self.model_card_kwargs.get("output_dir"):
            self.model_card_kwargs.update({"output_dir": f'"{output_dir}"'})

        model_card = ArgillaModelCard.from_template(
            card_data=self._trainer.get_model_card_data(**self.model_card_kwargs),
            template_path=ArgillaModelCard.default_template_path,
        )

        if output_dir:
            model_card_path = Path(output_dir) / "README.md"
            model_card.save(model_card_path)
            self._logger.info(f"Model card generated at: {model_card_path}")

        return model_card

    def push_to_huggingface(self, repo_id: str, generate_card: Optional[bool] = True, **kwargs) -> None:
        """Push your model to [huggingface's model hub](https://huggingface.co/models).

        Args:
            repo_id:
                The name of the repository you want to push your model and tokenizer to.
                It should contain your organization name when pushing to a given organization.
            generate_card:
                Whether to generate (and push) a model card for your model. Defaults to True.
        """
        if not kwargs.get("token"):
            # Try obtaining the token with huggingface_hub utils as a last resort, or let it fail.
            from huggingface_hub import HfFolder

            if token := HfFolder.get_token():
                kwargs["token"] = token

            # One last check for the tests. We use a different env var name
            # that the one gathered with HfFolder.get_token
            if token := kwargs.get("token", os.environ.get("HF_HUB_ACCESS_TOKEN", None)):
                kwargs["token"] = token

        url = self._trainer.push_to_huggingface(repo_id, **kwargs)

        if generate_card:
            model_card = self.generate_model_card()
            # For spacy based models, overwrite the repo_id with the url variable returned
            # from its trainer.
            if getattr(self._trainer, "language", None):
                repo_id = url

            model_card.push_to_hub(repo_id, repo_type="model", token=kwargs["token"])

    def update_config(self, *args, **kwargs) -> None:
        """
        Updates the `model_kwargs` and `trainer_kwargs` dictionaries with the keyword.add()

        Provides a warning if the keyword argument is not valid for the trainer or model.
        """

        def get_all_keys(d):
            keys = []
            for k, v in d.items():
                keys.append(k)
                if isinstance(v, dict):
                    keys += get_all_keys(v)
            return keys

        trainer_kwargs = self._trainer.get_trainer_kwargs()
        model_kwargs = self._trainer.get_model_kwargs()

        all_keys = get_all_keys({**trainer_kwargs, **model_kwargs})

        for kwarg in kwargs:
            if kwarg not in all_keys:
                warnings.warn(
                    f"'{kwarg}' is not a valid default argument for '{self._trainer.__class__.__name__}'. "
                    f"Valid default arguments are: {all_keys}. ",
                    UserWarning,
                    stacklevel=2,
                )
        return super().update_config(*args, **kwargs)


class ArgillaTrainerSkeleton(ArgillaTrainerSkeletonV1):
    def __init__(
        self,
        dataset: "FeedbackDataset",
        task: TrainingTaskTypes,
        prepared_data=None,
        model: str = None,
        seed: int = None,
        *arg,
        **kwargs,
    ):
        self._dataset = dataset
        self._task = task
        self._dataset = prepared_data
        self._model = model
        self._seed = seed
        if isinstance(self._task, TrainingTaskForTextClassification):
            self._multi_label = self._task.__multi_label__ or False
            self._label_list = self._task.__all_labels__ or None
            self._label2id = self._task.__label2id__
            self._id2label = self._task.__id2label__
            self._record_class = TextClassificationRecord  # TODO: dirty hack to inherit from original trainers
        else:
            self._record_class = FeedbackRecord
        self.model_kwargs = {}
        self.trainer_kwargs = {}
        self.trainer_model = None
        self.trainer_tokenizer = None
        self._trainer = None



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Dict, List

if TYPE_CHECKING:
    from argilla.client.feedback.schemas.questions import QuestionSchema


class LabelMappingMixin:
    """Mixin to provide label mapping functionality to a `QuestionSchema`."""

    @property
    def __all_labels__(self: "QuestionSchema") -> List[str]:
        return [entry["value"] for entry in self.server_settings["options"]]

    @property
    def __label2id__(self) -> Dict[str, int]:
        return {label: idx for idx, label in enumerate(self.__all_labels__)}

    @property
    def __id2label__(self) -> Dict[int, str]:
        return {idx: label for idx, label in enumerate(self.__all_labels__)}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

from argilla.client import singleton
from argilla.metrics import helpers
from argilla.metrics.models import MetricSummary


def f1(name: str, query: Optional[str] = None) -> MetricSummary:
    """Computes the single label f1 metric for a dataset

    Args:
        name:
            The dataset name.
        query:
            An ElasticSearch query with the [query string syntax](https://docs.argilla.io/en/latest/practical_guides/filter_dataset.html)

    Returns:
        The f1 metric summary

    Examples:
        >>> from argilla.metrics.text_classification import f1
        >>> summary = f1(name="example-dataset")
        >>> summary.visualize() # will plot a bar chart with results
        >>> summary.data # returns the raw result data
    """
    metric = singleton.active_api().compute_metric(name, metric="F1", query=query)

    return MetricSummary.new_summary(
        data=metric.results,
        visualization=lambda: helpers.f1(data=metric.results, title=metric.description),
    )


def f1_multilabel(name: str, query: Optional[str] = None) -> MetricSummary:
    """Computes the multi-label label f1 metric for a dataset

    Args:
        name:
            The dataset name.
        query:
            An ElasticSearch query with the [query string syntax](https://docs.argilla.io/en/latest/practical_guides/filter_dataset.html)

    Returns:
        The f1 metric summary

    Examples:
        >>> from argilla.metrics.text_classification import f1_multilabel
        >>> summary = f1_multilabel(name="example-dataset")
        >>> summary.visualize() # will plot a bar chart with results
        >>> summary.data # returns the raw result data
    """
    metric = singleton.active_api().compute_metric(name, metric="MultiLabelF1", query=query)

    return MetricSummary.new_summary(
        data=metric.results,
        visualization=lambda: helpers.f1(data=metric.results, title=metric.description),
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Any, Dict, Optional

if TYPE_CHECKING:
    from argilla.pydantic_v1 import BaseModel


def title_must_have_value(cls: "BaseModel", v: Optional[str], values: Dict[str, Any]) -> str:
    """Function to add as a `pydantic` validator to a field.

    This function will be called when the field is being validated, and it will
    receive the value of the field, and the values of the model being validated.

    If the field `title` is not empty, then it will be returned as is. Otherwise,
    the value of the field `name` will be capitalized and returned.
    """
    if not v:
        # If `name` doesn't pass the regex validation, then `values` won't have it
        name = values.get("name")
        if name:
            # If `name` has a value, then capitalize it and return it as the `title`
            return name.capitalize()
    return v


def validate_numeric_metadata_property_bounds(cls: "BaseModel", values: Dict[str, Any]) -> Dict[str, Any]:
    name = values.get("name")
    min = values.get("min")
    max = values.get("max")

    if min is not None and max is not None and min >= max:
        raise ValueError(f"`{cls.__name__}` with name={name} cannot have `max` less or equal than `min`")

    return values


def validate_numeric_metadata_filter_bounds(cls: "BaseModel", values: Dict[str, Any]) -> Dict[str, Any]:
    name = values.get("name")
    le = values.get("le")
    ge = values.get("ge")

    if le is None and ge is None:
        raise ValueError(f"`{cls.__name__}` with name={name} must have at least one of `le` or `ge`")

    if le is not None and ge is not None and le < ge:
        raise ValueError(f"`{cls.__name__}` with name={name} cannot have `le` less than `ge`")

    return values



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional, Union
from uuid import UUID

import httpx

from argilla.client.sdk.client import AuthenticatedClient
from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import ErrorMessage, HTTPValidationError, Response
from argilla.client.sdk.users.models import UserCreateModel, UserModel, UserRole


# TODO(alvarobartt,frascuchon): use ONLY `httpx.Client` instead of `AuthenticatedClient` and
# fix mock in `tests/conftest.py` to use `httpx.Client` instead of `AuthenticatedClient`
def whoami(client: AuthenticatedClient) -> UserModel:
    """Sends a GET request to `/api/me` endpoint to get the current user information.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.

    Returns:
        A `UserModel` instance with the current user information.
    """
    url = "/api/me"

    response = client.get(url)
    return UserModel(**response)


# TODO(frascuchon): rename this to `whoami` and deprecate the current `whoami` function
# in favor of this one, as this is just a patch.
def whoami_httpx(client: httpx.Client) -> Response[Union[UserModel, ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/me` endpoint to get the current user information.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is an instance of `UserModel`.
    """
    url = "/api/me"

    response = client.get(url)

    if response.status_code == 200:
        parsed_response = UserModel(**response.json())
        return Response(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
            parsed=parsed_response,
        )
    return handle_response_error(response)


def list_users(client: httpx.Client) -> Response[Union[List[UserModel], ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/users` endpoint to get the list of users.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a list of `UserModel` instances.
    """
    url = "/api/users"

    response = client.get(url=url)

    if response.status_code == 200:
        parsed_response = [UserModel(**user) for user in response.json()]
        return Response(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
            parsed=parsed_response,
        )
    return handle_response_error(response)


def create_user(
    client: httpx.Client,
    first_name: str,
    username: str,
    password: str,
    last_name: Optional[str] = None,
    role: UserRole = UserRole.annotator,
    workspaces: Optional[List[str]] = None,
) -> Response[Union[UserModel, ErrorMessage, HTTPValidationError]]:
    """Sends a POST request to `/api/users` endpoint to create a new user.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        first_name: the first name of the user. Must be a string of at least 1 character.
        last_name: the last name of the user. Is optional and defaults to `None`, but if provided
            must be a string of at least 1 character.
        username: the username of the user. Must be a string matching the following regex:
            ^(?!-|_)[a-z0-9-_]+$.
        role: the role of the user. Available roles are: `admin`, and `annotator`.
        password: the password of the user. Must be a string between 8 and 100 characters.
        workspaces: a list of workspace names to which the user will be linked to.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is an instance of `UserModel`.
    """
    url = "/api/users"

    user = UserCreateModel(
        first_name=first_name,
        last_name=last_name,
        username=username,
        role=role,
        password=password,
        workspaces=workspaces,
    )

    response = client.post(
        url=url,
        json=user.dict(exclude_none=True),
    )

    if response.status_code == 200:
        parsed_response = UserModel(**response.json())
        return Response(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
            parsed=parsed_response,
        )
    return handle_response_error(response)


def delete_user(client: httpx.Client, user_id: UUID) -> Response[Union[UserModel, ErrorMessage, HTTPValidationError]]:
    """Sends a DELETE request to `/api/users/{user_id}` endpoint to delete a user.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        user_id: the id of the user to be deleted.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is an instance of `UserModel`.
    """
    url = f"/api/users/{user_id}"

    response = client.delete(url=url)

    if response.status_code == 200:
        parsed_response = UserModel(**response.json())
        return Response(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
            parsed=parsed_response,
        )
    return handle_response_error(response)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import atexit
import logging
import random
import threading
from queue import Empty, Queue
from typing import Any, Dict, Iterable, List, Optional

import backoff
import monotonic
import wrapt

from argilla.client.api import Api
from argilla.client.models import Record
from argilla.client.sdk.commons.errors import ArApiResponseError


class ModelNotSupportedError(Exception):
    pass


class DatasetRecordsConsumer(threading.Thread):
    """Consumes the records from the dataset queue."""

    log = logging.getLogger("argilla.monitoring")

    def __init__(
        self,
        name: str,
        api: Api,
        tags: Optional[dict] = None,
        metadata: Optional[dict] = None,
        buffer_size: int = 10000,
        upload_size=256,
        upload_interval=1.0,
        retries=10,
        timeout=15,
        on_error=None,
    ):
        """Create a consumer thread."""
        threading.Thread.__init__(self)
        self.daemon = True
        self.upload_size = upload_size
        self.upload_interval = upload_interval
        self.api = api
        self.on_error = on_error
        self.queue = Queue(maxsize=buffer_size)
        self.dataset = name
        self.tags = tags
        self.metadata = metadata

        self.running = True
        self.retries = retries
        self.timeout = timeout

    def run(self):
        """Runs the consumer."""
        while self.running:
            self.log_next_batch()
        self._log_remaining()

    def _log_remaining(self):
        """When the consumer is paused, send remaining data in the queue"""
        last_batch = []
        while not self.queue.empty():
            last_batch.append(self.queue.get())
        if not len(last_batch):
            return

        self._log_records(last_batch)
        for _ in last_batch:
            self.queue.task_done()

    def pause(self):
        """Pause the consumer."""
        self.running = False

    def log_next_batch(self):
        """Upload the next batch of items, return whether successful."""

        batch = self._next_batch()
        if len(batch) == 0:
            return False

        success = self._log_records(batch)
        # mark items as acknowledged from queue
        for _ in batch:
            self.queue.task_done()
        return success

    def _next_batch(self) -> List[Record]:
        queue = self.queue
        records = []

        start_time = monotonic.monotonic()
        while len(records) < self.upload_size:
            elapsed = monotonic.monotonic() - start_time
            if elapsed >= self.upload_interval:
                break
            try:
                item = queue.get(
                    block=True,
                    timeout=self.upload_interval - elapsed,
                )
                records.append(item)
            except Empty:
                break

        return records

    def _log_records(self, batch: List[Record]):
        def fatal_exception(exc):
            if isinstance(exc, ArApiResponseError):
                return (400 <= exc.HTTP_STATUS < 500) and exc.HTTP_STATUS != 429
            else:
                return False

        @backoff.on_exception(
            backoff.expo,
            Exception,
            max_tries=self.retries + 1,
            giveup=fatal_exception,
        )
        def _inner_log_records(batch_: List[Record]):
            self.api.log(
                name=self.dataset,
                records=batch_,
                tags=self.tags,
                metadata=self.metadata,
                verbose=False,
            )

        try:
            _inner_log_records(batch)
            return True
        except Exception as e:
            self.log.error("error logging data: %s", e)
            if self.on_error:
                self.on_error(e, batch)
            return False

    def send(self, records: Iterable[Record]):
        """Send records to the consumer"""
        for record in records:
            self.queue.put(
                item=record,
                block=False,
            )


class BaseMonitor(wrapt.ObjectProxy):
    """
    A base monitor class for easy task model monitoring

    Attributes:
    -----------
    dataset:
        argilla dataset name
    sample_rate:
        The portion of the data to be store in argilla. Default = 0.2
    """

    def __init__(
        self,
        *args,
        api: Api,
        dataset: str,
        sample_rate: float = 1.0,
        log_interval: float = 1.0,
        agent: Optional[str] = None,
        tags: Dict[str, str] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        assert dataset, "Missing dataset"
        assert 0.0 < sample_rate <= 1.0, "Wrong sample rate. Set a value in (0, 1] range."

        self.dataset = dataset
        self.sample_rate = sample_rate
        self.agent = agent
        self.tags = tags
        self._api = api
        self._log_interval = log_interval
        self._consumers: Dict[str, DatasetRecordsConsumer] = {}

        atexit.register(self.shutdown)

    @property
    def __model__(self):
        """Return the monitored task model"""
        return self.__wrapped__

    def is_record_accepted(self) -> bool:
        """Return True if a record should be logged to argilla"""
        return random.uniform(0.0, 1.0) <= self.sample_rate

    def _prepare_log_data(self, *args, **kwargs) -> Dict[str, Any]:
        raise NotImplementedError()

    def shutdown(self):
        """Stop consumers"""
        for consumer in self._consumers.values():
            try:
                consumer.pause()
                consumer.join()
            except RuntimeError:
                pass

    def send_records(self, *args, **kwargs):
        data = self._prepare_log_data(*args, **kwargs)

        consumer = self._get_consumer_by_dataset(dataset=data["name"])
        consumer.tags = data.get("tags", {})
        consumer.metadata = data.get("metadata", {})
        consumer.send(data["records"])

    def _get_consumer_by_dataset(self, dataset: str):
        if dataset not in self._consumers:
            self._consumers[dataset] = self._create_consumer(dataset)
        return self._consumers[dataset]

    def _create_consumer(self, name: str):
        consumer = DatasetRecordsConsumer(
            name=name,
            api=self._api,
            upload_interval=self._log_interval,
        )
        consumer.start()
        return consumer



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import textwrap
import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Literal, Optional, Tuple, Union

from rich.progress import Progress

from argilla.client.feedback.constants import DELETE_DATASET_RECORDS_MAX_NUMBER, PUSHING_BATCH_SIZE
from argilla.client.feedback.dataset import helpers
from argilla.client.feedback.dataset.base import FeedbackDatasetBase, SortBy
from argilla.client.feedback.dataset.mixins import MetricsMixin, UnificationMixin
from argilla.client.feedback.dataset.remote.mixins import ArgillaRecordsMixin
from argilla.client.feedback.mixins import ArgillaMetadataPropertiesMixin
from argilla.client.feedback.schemas.enums import ResponseStatusFilter
from argilla.client.feedback.schemas.records import FeedbackRecord
from argilla.client.feedback.schemas.remote.records import RemoteFeedbackRecord
from argilla.client.feedback.schemas.remote.vector_settings import RemoteVectorSettings
from argilla.client.feedback.schemas.vector_settings import VectorSettings
from argilla.client.feedback.training.schemas.base import (
    TrainingTaskTypes,
)
from argilla.client.models import Framework
from argilla.client.sdk.commons.errors import AlreadyExistsApiError
from argilla.client.sdk.users.models import UserRole
from argilla.client.sdk.v1.datasets import api as datasets_api_v1
from argilla.client.sdk.v1.datasets.models import FeedbackRecordsSearchVectorQuery
from argilla.client.sdk.v1.vectors_settings import api as vectors_settings_api_v1
from argilla.client.utils import allowed_for_roles

INCLUDE_ALL_VECTORS_PARAM: str = "all"

if TYPE_CHECKING:
    from uuid import UUID

    import httpx

    from argilla.client.feedback.dataset.local.dataset import FeedbackDataset
    from argilla.client.feedback.schemas.metadata import MetadataFilters
    from argilla.client.feedback.schemas.types import (
        AllowedMetadataPropertyTypes,
        AllowedRemoteFieldTypes,
        AllowedRemoteMetadataPropertyTypes,
        AllowedRemoteQuestionTypes,
    )
    from argilla.client.sdk.v1.datasets.models import FeedbackRecordsModel
    from argilla.client.workspaces import Workspace


class RemoteFeedbackRecords(ArgillaRecordsMixin):
    def __init__(
        self,
        dataset: "RemoteFeedbackDataset",
        response_status: Optional[Union[ResponseStatusFilter, List[ResponseStatusFilter]]] = None,
        metadata_filters: Optional[Union["MetadataFilters", List["MetadataFilters"]]] = None,
        sort_by: Optional[List[SortBy]] = None,
        with_vectors: Union[INCLUDE_ALL_VECTORS_PARAM, List[str], None] = None,
    ) -> None:
        """Initializes a `RemoteFeedbackRecords` instance to access a `FeedbackDataset`
        records in Argilla. This class is used to get records from Argilla, iterate over
        them, and push new records to Argilla.

        Note:
            This class is not intended to be initialised directly. Instead, use
            `FeedbackDataset.from_argilla` to get an instance of `RemoteFeedbackDataset`,
            and then just call `records` on it.

        Args:
            dataset: the `RemoteFeedbackDataset` instance to access the `httpx.Client`,
                the ID of the dataset in Argilla, and everything else to reuse some methods
                and/or attributes.
        """
        self._dataset = dataset

        if response_status and not isinstance(response_status, list):
            response_status = [response_status]
        if metadata_filters and not isinstance(metadata_filters, list):
            metadata_filters = [metadata_filters]

        # TODO: Validate filters and sort exists in metadata
        self._sort_by = sort_by or []
        self._response_status = response_status or []
        self._metadata_filters = metadata_filters or []

        self._with_vectors = with_vectors

        self._validate_metadata_names()
        self._validate_vector_names()

    @property
    def dataset(self) -> "RemoteFeedbackDataset":
        """Returns the `RemoteFeedbackDataset` instance that this `RemoteFeedbackRecords` belongs to."""
        return self._dataset

    @property
    def sort_by(self) -> Optional[List[SortBy]]:
        """Returns the sort by fields and orders that this `RemoteFeedbackRecords` is using."""
        return [sort_by for sort_by in self._sort_by]

    @property
    def metadata_filters(self) -> Optional[List["MetadataFilters"]]:
        """Returns the metadata filters that this `RemoteFeedbackRecords` is using."""
        return [metadata_filter for metadata_filter in self._metadata_filters]

    @property
    def response_status(self) -> Optional[List[ResponseStatusFilter]]:
        """Returns the response status filters that this `RemoteFeedbackRecords` is using."""
        return [ResponseStatusFilter(response_status) for response_status in self._response_status]

    @property
    def _client(self) -> "httpx.Client":
        """Returns the `httpx.Client` instance that will be used to send requests to Argilla."""
        return self.dataset._client

    @property
    def _question_id_to_name(self) -> Dict["UUID", str]:
        return self.dataset._question_id_to_name

    @property
    def _question_name_to_id(self) -> Dict[str, "UUID"]:
        return self.dataset._question_name_to_id

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def __len__(self) -> int:
        """Returns the number of records in the current `FeedbackDataset` in Argilla."""
        try:
            if self._has_filters():
                return self._fetch_records(offset=0, limit=1).total
            else:
                response = datasets_api_v1.get_metrics(client=self._client, id=self._dataset.id).parsed
                return response.records.count
        except Exception as e:
            raise Exception(
                f"Failed while getting the metrics from the current `FeedbackDataset` in Argilla with exception: {e}"
            ) from e

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def add(
        self,
        records: Union[FeedbackRecord, Dict[str, Any], List[Union[FeedbackRecord, Dict[str, Any]]]],
        show_progress: bool = True,
    ) -> None:
        """Pushes a list of `FeedbackRecord`s to Argilla.

        Args:
            records: can be a single `FeedbackRecord`, a list of `FeedbackRecord`,
                a single dictionary, or a list of dictionaries. If a dictionary is provided,
                it will be converted to a `FeedbackRecord` internally.
            show_progress: Whether to show a `tqdm` progress bar while pushing the records.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            Exception: If the pushing of the records to Argilla fails.
        """
        records = helpers.normalize_records(records)
        helpers.validate_dataset_records(self.dataset, records)

        question_name_to_id = {question.name: question.id for question in self.dataset.questions}

        with Progress() as progress_bar:
            task = progress_bar.add_task("Pushing records to Argilla...", total=len(records), visible=show_progress)

            for i in range(0, len(records), PUSHING_BATCH_SIZE):
                batch = records[i : i + PUSHING_BATCH_SIZE]
                datasets_api_v1.add_records(
                    client=self._client,
                    id=self.dataset.id,
                    records=[record.to_server_payload(question_name_to_id=question_name_to_id) for record in batch],
                )
                progress_bar.update(task, advance=len(batch))

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def update(
        self,
        records: Union[RemoteFeedbackRecord, List[RemoteFeedbackRecord]],
        show_progress: bool = True,
    ) -> None:
        """Updates a list of `RemoteFeedbackRecord`s in Argilla.

        Args:
            records: can be a single `RemoteFeedbackRecord` or a list of
                `RemoteFeedbackRecord`.
            show_progress: Whether to show a `tqdm` progress bar while updating the records.
        """
        if isinstance(records, RemoteFeedbackRecord):
            records = [records]

        helpers.validate_dataset_records(self.dataset, records, attributes_to_validate=["metadata", "vectors"])

        with Progress() as progress_bar:
            task = progress_bar.add_task("Updating records in Argilla...", total=len(records), visible=show_progress)

            for i in range(0, len(records), PUSHING_BATCH_SIZE):
                batch = records[i : i + PUSHING_BATCH_SIZE]
                datasets_api_v1.update_records(
                    client=self._client,
                    id=self.dataset.id,
                    records=[
                        {"id": str(record.id), **record.to_server_payload(self._question_name_to_id)}
                        for record in batch
                    ],
                )
                progress_bar.update(task, advance=len(batch))

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def delete(self, records: List[RemoteFeedbackRecord]) -> None:
        """Deletes a list of `RemoteFeedbackRecord`s from Argilla.

        Args:
            records: A list of `RemoteFeedbackRecord`s to delete from Argilla.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: If the deletion of the records from Argilla fails.
        """
        num_records = len(records)
        for start in range(0, num_records, DELETE_DATASET_RECORDS_MAX_NUMBER):
            end = min(start + DELETE_DATASET_RECORDS_MAX_NUMBER, num_records)
            try:
                datasets_api_v1.delete_records(
                    client=self._client,
                    id=self._dataset.id,
                    record_ids=[record.id for record in records[start:end]],
                )
            except Exception as e:
                raise RuntimeError("Failed to remove records from Argilla") from e

    def _fetch_records(self, offset: int, limit: int) -> "FeedbackRecordsModel":
        """Fetches a batch of records from Argilla."""

        return datasets_api_v1.get_records(
            client=self._client,
            id=self._dataset.id,
            include=self.include_as_query_params,
            offset=offset,
            limit=limit,
            response_status=self.response_status_as_query_string,
            metadata_filters=self.metadata_filters_as_query_strings,
            sort_by=self.__sort_by_query_strings,
        ).parsed

    def _has_filters(self) -> bool:
        """Returns whether the current `RemoteFeedbackRecords` is filtered or not."""
        return bool(self._response_status) or bool(self._metadata_filters)

    # TODO: define `List[ResponseStatusFilter]` and delegate `query_string` formatting to it
    @property
    def response_status_as_query_string(self) -> Optional[List[str]]:
        """Formats the `response_status` if any to the query string format. Otherwise, returns `None`."""
        return (
            [status.value if hasattr(status, "value") else status for status in self._response_status]
            if len(self._response_status) > 0
            else None
        )

    # TODO: define `List[MetadataFilter]` and delegate `query_string` formatting to it
    @property
    def metadata_filters_as_query_strings(self) -> Optional[List[str]]:
        """Formats the `metadata_filters` if any to the query string format. Otherwise, returns `None`."""
        return (
            [metadata_filter.query_string for metadata_filter in self._metadata_filters]
            if len(self._metadata_filters) > 0
            else None
        )

    # TODO: define `List[SortBy]` and delegate `query_string` formatting to it
    @property
    def __sort_by_query_strings(self) -> Optional[List[str]]:
        """Formats the `sort_by` if any to the query string format. Otherwise, returns `None`."""
        return [f"{sort_by.field}:{sort_by.order}" for sort_by in self._sort_by] if len(self._sort_by) > 0 else None

    @classmethod
    def _create_from_dataset(
        cls,
        new_ds: "RemoteFeedbackDataset",
        response_status: Optional[Union[ResponseStatusFilter, List[ResponseStatusFilter]]] = None,
        metadata_filters: Optional[Union["MetadataFilters", List["MetadataFilters"]]] = None,
        sort_by: Optional[List[SortBy]] = None,
    ):
        """Creates a new instance of `RemoteFeedbackRecords` with the given filters."""

        sort_by = sort_by or new_ds.records.sort_by
        metadata_filters = metadata_filters or new_ds.records.metadata_filters
        response_status = response_status or new_ds.records.response_status

        return cls(
            new_ds,
            sort_by=sort_by,
            metadata_filters=metadata_filters,
            response_status=response_status,
            with_vectors=new_ds.records._with_vectors,
        )

    def _validate_metadata_names(self):
        names = []
        if self.metadata_filters:
            names.extend([metadata_filter.name for metadata_filter in self.metadata_filters])
        if self.sort_by:
            names.extend([sort.metadata_name for sort in self.sort_by if sort.is_metadata_field])
        if names:
            helpers.validate_metadata_names(self.dataset, names)

    def _validate_vector_names(self):
        if not self._with_vectors or self._with_vectors == INCLUDE_ALL_VECTORS_PARAM:
            return

        if isinstance(self._with_vectors, str):
            self._with_vectors = [self._with_vectors]

        helpers.validate_vector_names(self.dataset, self._with_vectors)

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def find_similar(
        self,
        vector_name: str,
        value: Optional[List[float]] = None,
        record: Optional[RemoteFeedbackRecord] = None,
        max_results: int = 50,
    ) -> List[Tuple[RemoteFeedbackRecord, float]]:
        if bool(record) == bool(value):
            raise ValueError("Either 'record' or 'value' must be provided")

        try:
            response = datasets_api_v1.search_records(
                client=self._client,
                id=self.dataset.id,
                include=self.include_as_query_params,
                vector_query=FeedbackRecordsSearchVectorQuery(
                    name=vector_name,
                    record_id=record and record.id,
                    value=value,
                ),
                metadata_filters=self.metadata_filters_as_query_strings,
                response_status=self.response_status_as_query_string,
                limit=max_results,
            )
        except Exception as e:
            raise RuntimeError(f"Failed searching records for dataset with exception: {e}") from e

        question_id_to_name_map = self.dataset._question_id_to_name
        return [
            (
                RemoteFeedbackRecord.from_api(record_score.record, question_id_to_name=question_id_to_name_map),
                record_score.query_score,
            )
            for record_score in response.parsed.items
        ]

    @property
    def include_as_query_params(self) -> List[str]:
        include = ["responses", "suggestions"]  # default include

        if not self._with_vectors:
            return include

        if self._with_vectors == INCLUDE_ALL_VECTORS_PARAM:
            include.append("vectors")
        else:
            include.append(f"vectors:{','.join(self._with_vectors)}")

        return include


class RemoteFeedbackDataset(FeedbackDatasetBase[RemoteFeedbackRecord], MetricsMixin, UnificationMixin):
    # TODO: Call super method once the base init contains only commons init attributes
    def __init__(
        self,
        *,
        client: "httpx.Client",
        id: "UUID",
        name: str,
        workspace: "Workspace",
        created_at: datetime,
        updated_at: datetime,
        fields: List["AllowedRemoteFieldTypes"],
        questions: List["AllowedRemoteQuestionTypes"],
        guidelines: Optional[str] = None,
        allow_extra_metadata: bool = True,
        with_vectors: Union[Literal[INCLUDE_ALL_VECTORS_PARAM], List[str], None] = None,
    ) -> None:
        """Initializes a `RemoteFeedbackDataset` instance in Argilla.

        Note:
            This class is not intended to be initialised directly. Instead, use
            `FeedbackDataset.from_argilla` to get an instance of this class.

        Args:
            client: contains the `httpx.Client` instance that will be used to send requests to Argilla.
            id: contains the UUID of the dataset in Argilla.
            name: contains the name of the dataset in Argilla.
            workspace: contains the `Workspace` instance that the dataset belongs to in Argilla.
            created_at: contains the datetime when the dataset was created in Argilla.
            updated_at: contains the datetime when the dataset was last updated in Argilla.
            fields: contains the fields that will define the schema of the records in the dataset.
            questions: contains the questions that will be used to annotate the dataset.
            guidelines: contains the guidelines for annotating the dataset. Defaults to `None`.

        Raises:
            TypeError: if `fields` is not a list of `FieldSchema`.
            ValueError: if `fields` does not contain at least one required field.
            TypeError: if `questions` is not a list of `TextQuestion`, `RatingQuestion`,
                `LabelQuestion`, and/or `MultiLabelQuestion`.
            ValueError: if `questions` does not contain at least one required question.
            TypeError: if `guidelines` is not None and not a string.
            ValueError: if `guidelines` is an empty string.
        """

        self._fields = fields
        self._questions = questions
        self._guidelines = guidelines
        self._allow_extra_metadata = allow_extra_metadata

        self._client = client  # Required to be able to use `allowed_for_roles` decorator
        self._id = id
        self._name = name
        self._workspace = workspace
        self._created_at = created_at
        self._updated_at = updated_at

        self._records = RemoteFeedbackRecords(dataset=self, with_vectors=with_vectors)

    @property
    def guidelines(self) -> Optional[str]:
        return self._guidelines

    @property
    def allow_extra_metadata(self) -> bool:
        return self._allow_extra_metadata

    @property
    def fields(self) -> Union[List["AllowedRemoteFieldTypes"]]:
        return self._fields

    @property
    def questions(self) -> Union[List["AllowedRemoteQuestionTypes"]]:
        return self._questions

    @property
    def records(self) -> RemoteFeedbackRecords:
        """Returns an instance of `RemoteFeedbackRecords` that allows you to iterate over
        the records in the dataset. The records are fetched from Argilla on the fly and
        not stored in memory. You can also iterate over the records directly from the
        dataset instance.
        """
        return self._records

    @property
    def id(self) -> "UUID":
        """Returns the ID of the dataset in Argilla."""
        return self._id

    @property
    def name(self) -> str:
        """Returns the name of the dataset in Argilla."""
        return self._name

    @property
    def workspace(self) -> "Workspace":
        """Returns the workspace the dataset belongs to in Argilla."""
        return self._workspace

    @property
    def url(self) -> str:
        """Returns the URL of the dataset in Argilla."""
        return f"{self._client.base_url}/dataset/{self.id}/annotation-mode"

    @property
    def created_at(self) -> datetime:
        """Returns the datetime when the dataset was created in Argilla."""
        return self._created_at

    @property
    def updated_at(self) -> datetime:
        """Returns the datetime when the dataset was last updated in Argilla."""
        return self._updated_at

    @property
    def _question_id_to_name(self) -> Dict["UUID", str]:
        return {question.id: question.name for question in self._questions}

    @property
    def _question_name_to_id(self) -> Dict[str, "UUID"]:
        return {question.name: question.id for question in self._questions}

    @property
    def metadata_properties(self) -> List["AllowedRemoteMetadataPropertyTypes"]:
        """Retrieves the `metadata_properties` of the current dataset from Argilla, and
        returns them if any, otherwise, it returns an empty list.
        """
        return ArgillaMetadataPropertiesMixin.list(client=self._client, dataset_id=self.id)

    @property
    def vectors_settings(self) -> List[RemoteVectorSettings]:
        """Retrieves the `vectors_settings` of the current dataset from Argilla"""
        response = datasets_api_v1.list_vectors_settings(client=self._client, id=self.id)

        return [RemoteVectorSettings.from_api(vector_settings) for vector_settings in response.parsed.items]

    def __repr__(self) -> str:
        """Returns a string representation of the dataset."""
        indent = "   "
        return (
            "RemoteFeedbackDataset("
            + textwrap.indent(f"\nid={self.id}", indent)
            + textwrap.indent(f"\nname={self.name}", indent)
            + textwrap.indent(f"\nworkspace={self.workspace}", indent)
            + textwrap.indent(f"\nurl={self.url}", indent)
            + textwrap.indent(f"\nfields={self.fields}", indent)
            + textwrap.indent(f"\nquestions={self.questions}", indent)
            + textwrap.indent(f"\nguidelines={self.guidelines}", indent)
            + textwrap.indent(f"\nmetadata_properties={self.metadata_properties}", indent)
            + textwrap.indent(f"\nvectors_settings={self.vectors_settings}", indent)
            + "\n)"
        )

    def __len__(self) -> int:
        """Returns the number of records in the dataset."""
        return self._records.__len__()

    def __iter__(self) -> Iterator[RemoteFeedbackRecord]:
        """Returns an iterator over the records in the dataset."""
        yield from self._records

    def __getitem__(self, key: Union[slice, int]) -> Union[RemoteFeedbackRecord, List[RemoteFeedbackRecord]]:
        """Returns the record(s) at the given index(es).

        Args:
            key: The index or slice to retrieve.

        Returns:
            The record(s) at the given index(es).
        """
        return self._records.__getitem__(key)

    def sort_by(self, sort: List[SortBy]) -> "RemoteFeedbackDataset":
        """Sorts the current `RemoteFeedbackDataset` based on the given sort fields and orders."""
        sorted_dataset = self._create_from_dataset(self)
        sorted_dataset._records = RemoteFeedbackRecords._create_from_dataset(sorted_dataset, sort_by=sort)

        return sorted_dataset

    def add_records(
        self,
        records: Union["FeedbackRecord", Dict[str, Any], List[Union["FeedbackRecord", Dict[str, Any]]]],
        show_progress: bool = True,
    ) -> None:
        """Adds the given records to the dataset and pushes those to Argilla.

        Args:
            records: can be a single `FeedbackRecord`, a list of `FeedbackRecord`,
                a single dictionary, or a list of dictionaries. If a dictionary is provided,
                it will be converted to a `FeedbackRecord` internally.
            show_progress: if `True`, shows a progress bar while pushing the records to
                Argilla. Defaults to `True`.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            ValueError: if the given records are neither: `FeedbackRecord`, list of
                `FeedbackRecord`, list of dictionaries as a record or dictionary as a
                record; or if the given records do not match the expected schema.
        """
        self._records.add(records=records, show_progress=show_progress)

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def find_similar_records(
        self,
        vector_name: str,
        value: Optional[List[float]] = None,
        record: Optional[RemoteFeedbackRecord] = None,
        max_results: int = 50,
    ) -> List[Tuple[RemoteFeedbackRecord, float]]:
        """Finds similar records to the given record in the dataset based on the given vector.

        Args:
            vector_name: a vector name to use for searching by similarity.
            value: an optional vector value to be used for searching by similarity. Defaults to None.
            record: an optional record to be used for searching by similarity. Defaults to None.
            max_results: the maximum number of results for the search. Defaults to 50.

        Returns:
            A list of tuples with each tuple including a record and a similarity score.
        """
        return self.records.find_similar(vector_name, value, record, max_results)

    def update_records(
        self,
        records: Union[RemoteFeedbackRecord, List[RemoteFeedbackRecord]],
        show_progress: bool = True,
    ) -> None:
        """Updates the given records in the dataset in Argilla.

        Args:
            records: the records to update in the dataset. Can be a single record or a
                list of records. The records need to be previously pushed to Argilla,
                otherwise they won't be updated.
            show_progress: if `True`, shows a progress bar while pushing the records to
                Argilla. Defaults to `True`.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
        """
        self._records.update(records=records, show_progress=show_progress)

    def delete_records(self, records: Union["RemoteFeedbackRecord", List["RemoteFeedbackRecord"]]) -> None:
        """Deletes the given records from the dataset in Argilla.

        Args:
            records: the records to delete from the dataset. Can be a single record or a list
                of records. But those need to be previously pushed to Argilla, otherwise
                they won't be deleted.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: If the deletion of the records from Argilla fails.
        """
        self._records.delete(records=[records] if not isinstance(records, list) else records)

    def pull(self, max_records: Optional[int] = None) -> "FeedbackDataset":
        """Pulls the dataset from Argilla and returns a local instance of it.

        Args:
            max_records: the maximum number of records to pull from Argilla. Defaults to `None`.

        Returns:
            A local instance of the dataset which is a `FeedbackDataset` object.
        """
        # Importing here to avoid circular imports
        from argilla.client.feedback.dataset.local.dataset import FeedbackDataset

        instance = FeedbackDataset(
            fields=[field.to_local() for field in self.fields],
            questions=[question.to_local() for question in self.questions],
            guidelines=self.guidelines or None,
            metadata_properties=[metadata_property.to_local() for metadata_property in self.metadata_properties]
            or None,
            vectors_settings=[vector_settings.to_local() for vector_settings in self.vectors_settings] or None,
            allow_extra_metadata=self._allow_extra_metadata,
        )

        len_records = len(self._records)
        if len_records > 0:
            max_records = max_records or len_records
            records = [record.to_local() for record in self._records[:max_records]]
            instance.add_records(records=records)
        else:
            warnings.warn(
                "The dataset is empty, so no records will be added to the local instance.",
                UserWarning,
                stacklevel=1,
            )

        return instance

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def add_metadata_property(
        self, metadata_property: "AllowedMetadataPropertyTypes"
    ) -> "AllowedRemoteMetadataPropertyTypes":
        """Adds a new `metadata_property` to the current `FeedbackDataset` in Argilla.

        Note:
            Existing `FeedbackRecord`s if any will remain unchanged if those contain metadata
            named the same way as the `metadata_property`, but added before the
            `metadata_property` was added.

        Args:
            metadata_property: the metadata property to add to the current `FeedbackDataset`
                in Argilla.

        Returns:
            The newly added `metadata_property` to the current `FeedbackDataset` in Argilla.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: if the `metadata_property` cannot be added to the current
                `FeedbackDataset` in Argilla.
        """
        self._unique_metadata_property(metadata_property=metadata_property)

        try:
            metadata_property = datasets_api_v1.add_metadata_property(
                client=self._client,
                id=self.id,
                metadata_property=metadata_property.to_server_payload(),
            ).parsed
        except Exception as e:
            raise RuntimeError(
                f"Failed while adding the `metadata_property={metadata_property}` to the current `FeedbackDataset` in Argilla with exception: {e}"
            ) from e

        # TODO(alvarobartt): structure better the mixins to be able to easily reuse those, here to avoid circular imports
        from argilla.client.feedback.dataset.local.mixins import ArgillaMixin

        return ArgillaMixin._parse_to_remote_metadata_property(metadata_property=metadata_property, client=self._client)

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def update_metadata_properties(
        self,
        metadata_properties: Union["AllowedRemoteMetadataPropertyTypes", List["AllowedRemoteMetadataPropertyTypes"]],
    ) -> None:
        """Updates a list of `metadata_properties` in the current `FeedbackDataset` in Argilla.

        Note:
            All the `metadata_properties` provided must exist in Argilla in advance, and
            those will be pushed again to Argilla with the current values that they have,
            which ideally, should have been updated via assignment e.g. `metadata_property.title = \"...\"`.

        Args:
            metadata_properties: the metadata property/ies to update in the current `FeedbackDataset` in Argilla.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: if the `metadata_properties` cannot be updated in the current
                `FeedbackDataset` in Argilla.
        """
        if not isinstance(metadata_properties, list):
            metadata_properties = [metadata_properties]

        for metadata_property in metadata_properties:
            try:
                ArgillaMetadataPropertiesMixin.update(
                    client=self._client,
                    metadata_property_id=metadata_property.id,
                    title=metadata_property.title,
                    visible_for_annotators=metadata_property.visible_for_annotators,
                )
            except Exception as e:
                raise RuntimeError(
                    f"Failed while updating the `metadata_property={metadata_property.name}` in the current"
                    f" `FeedbackDataset` in Argilla with exception: {e}"
                ) from e

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def delete_metadata_properties(
        self, metadata_properties: Union[str, List[str]]
    ) -> Union["AllowedMetadataPropertyTypes", List["AllowedMetadataPropertyTypes"]]:
        """Deletes a list of `metadata_properties` from the current `FeedbackDataset`
        in Argilla.

        Note:
            Existing `FeedbackRecord`s if any, will remain unchanged if those contain metadata
            named the same way as the `metadata_properties` to delete, but the validation will
            be removed as well as `metadata_property` index, which means one won't be able to
            use that for filtering.

        Args:
            metadata_properties: the metadata property/ies name/s to delete from the current
                `FeedbackDataset` in Argilla.

        Returns:
            The `metadata_property` or `metadata_properties` deleted from the current
            `FeedbackDataset` in Argilla, but using the local schema e.g. if you delete a
            `RemoteFloatMetadataProperty` this method will delete it from Argilla and will
            return a `FloatMetadataProperty` instance.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: if the `metadata_properties` cannot be deleted from the current
                `FeedbackDataset` in Argilla.
        """
        if isinstance(metadata_properties, str):
            metadata_properties = [metadata_properties]

        existing_metadata_properties = self.metadata_properties
        existing_metadata_property_names = [
            metadata_property.name for metadata_property in existing_metadata_properties
        ]

        unexisting_metadata_properties = []
        for metadata_property in metadata_properties:
            if metadata_property not in existing_metadata_property_names:
                unexisting_metadata_properties.append(metadata_property)
        if len(unexisting_metadata_properties) > 0:
            raise ValueError(
                f"The following metadata properties do not exist in the current `FeedbackDataset` in Argilla: {unexisting_metadata_properties}."
                f" The existing metadata properties are: {existing_metadata_property_names}."
            )

        deleted_metadata_properties = []
        for metadata_property in existing_metadata_properties:
            if metadata_property.name in metadata_properties:
                ArgillaMetadataPropertiesMixin.delete(client=self._client, metadata_property_id=metadata_property.id)
                metadata_properties.remove(metadata_property.name)
                deleted_metadata_properties.append(metadata_property)
        return deleted_metadata_properties if len(deleted_metadata_properties) > 1 else deleted_metadata_properties[0]

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def add_vector_settings(self, vector_settings: VectorSettings) -> RemoteVectorSettings:
        """Adds a new vector settings to the current `FeedbackDataset` in Argilla.

        Args:
            vector_settings: the vector settings to add.

        Returns:
            The newly added vector settings to the current `FeedbackDataset` in Argilla.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            ValueError: if the vector settings with the given name already exists in the
                dataset in Argilla.
        """

        try:
            new_vector_settings = datasets_api_v1.add_vector_settings(
                client=self._client,
                id=self.id,
                title=vector_settings.title,
                name=vector_settings.name,
                dimensions=vector_settings.dimensions,
            ).parsed
        except AlreadyExistsApiError:
            raise ValueError(f"Vector settings with name {vector_settings.name!r} already exists.")
        return RemoteVectorSettings.from_api(new_vector_settings)

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def update_vectors_settings(
        self, vectors_settings: Union[RemoteVectorSettings, List[RemoteVectorSettings]]
    ) -> None:
        """Updates the given vector settings in the current `FeedbackDataset` in Argilla.

        Args:
            vectors_settings: the remote vectors settings to update. Must exist in Argilla in advance.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: if the vector settings cannot be updated in the current
                `FeedbackDataset` in Argilla.
        """
        if isinstance(vectors_settings, RemoteVectorSettings):
            vectors_settings = [vectors_settings]

        for vector_settings in vectors_settings:
            try:
                vectors_settings_api_v1.update_vector_settings(
                    client=self._client,
                    id=vector_settings.id,
                    title=vector_settings.title,
                ).parsed
            except Exception as e:
                raise RuntimeError(
                    f"Failed while updating the `vector_settings={vector_settings.name}` in the current"
                    f" `FeedbackDataset` in Argilla with exception: {e}"
                ) from e

    def delete_vectors_settings(
        self, vectors_settings: Union[str, List[str]]
    ) -> Union[RemoteVectorSettings, List["RemoteVectorSettings"]]:
        """Deletes the given vectors settings from the current `FeedbackDataset` in Argilla.

        Args:
            vectors_settings: the name/s of the vectors settings to delete.

        Returns:
            The vectors settings deleted from the current `FeedbackDataset` in Argilla.

        Raises:
            ValueError: if the given vectors settings do not exist in the current
                `FeedbackDataset` in Argilla.
        """
        if isinstance(vectors_settings, str):
            vectors_settings = [vectors_settings]

        existing_vectors_settings_name = [vector_settings.name for vector_settings in self.vectors_settings]

        unexisting_vectors_settings = []
        for vector_settings in vectors_settings:
            if vector_settings not in existing_vectors_settings_name:
                unexisting_vectors_settings.append(vector_settings.name)
        if len(unexisting_vectors_settings) > 0:
            raise ValueError(
                f"The following vectors settings do not exist in the current `FeedbackDataset` in Argilla: {unexisting_vectors_settings}."
                f" The existing vectors settings are: {existing_vectors_settings_name}."
            )

        deleted_vectors_settings = []
        for vector_settings in self.vectors_settings:
            if vector_settings.name in vectors_settings:
                vectors_settings_api_v1.delete_vector_settings(client=self._client, id=vector_settings.id).parsed
                vectors_settings.remove(vector_settings.name)
                deleted_vectors_settings.append(vector_settings)
        return deleted_vectors_settings if len(deleted_vectors_settings) > 1 else deleted_vectors_settings[0]

    def filter_by(
        self,
        *,
        response_status: Optional[Union[ResponseStatusFilter, List[ResponseStatusFilter]]] = None,
        metadata_filters: Optional[Union["MetadataFilters", List["MetadataFilters"]]] = None,
    ) -> "RemoteFeedbackDataset":
        """Filters the current `RemoteFeedbackDataset` based on the `response_status` of
        the responses of the records in Argilla. This method creates a new class instance
        of `FilteredRemoteFeedbackDataset` with the given filters.

        Args:
            response_status: the response status/es to filter the dataset by. Can be
                one of: draft, pending, submitted, and discarded. Defaults to `None`.
            metadata_filters: the metadata filters to filter the dataset by. Can be
                one of: `TermsMetadataFilter`, `IntegerMetadataFilter`, and
                `FloatMetadataFilter`. Defaults to `None`.

        Returns:
            A new instance of `FilteredRemoteFeedbackDataset` with the given filters.
        """
        if not response_status and not metadata_filters:
            raise ValueError("At least one of `response_status` or `metadata_filters` must be provided.")

        filtered_dataset = RemoteFeedbackDataset._create_from_dataset(self)
        filtered_dataset._records = RemoteFeedbackRecords._create_from_dataset(
            filtered_dataset, response_status=response_status, metadata_filters=metadata_filters
        )

        return filtered_dataset

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def delete(self) -> None:
        """Deletes the current `FeedbackDataset` from Argilla. This method is just working
        if the user has either `owner` or `admin` role.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: if the `FeedbackDataset` cannot be deleted from Argilla.
        """
        try:
            datasets_api_v1.delete_dataset(client=self._client, id=self.id)
        except Exception as e:
            raise RuntimeError(f"Failed while deleting the `FeedbackDataset` from Argilla with exception: {e}") from e

    @classmethod
    def _create_from_dataset(cls, dataset: "RemoteFeedbackDataset") -> "RemoteFeedbackDataset":
        new_dataset = cls(
            client=dataset._client,
            id=dataset.id,
            name=dataset.name,
            workspace=dataset.workspace,
            created_at=dataset.created_at,
            updated_at=dataset.updated_at,
            fields=dataset.fields,
            questions=dataset.questions,
            guidelines=dataset.guidelines,
        )

        new_dataset._records = dataset.records

        return new_dataset

    def prepare_for_training(
        self,
        framework: Union[Framework, str],
        task: TrainingTaskTypes,
        train_size: Optional[float] = 1,
        test_size: Optional[float] = None,
        seed: Optional[int] = None,
        lang: Optional[str] = None,
    ) -> Any:
        """
        Prepares the dataset for training for a specific training framework and NLP task by splitting the dataset into train and test sets.

        Args:
            framework: the framework to use for training. Currently supported frameworks are: `transformers`, `peft`,
                `setfit`, `spacy`, `spacy-transformers`, `span_marker`, `spark-nlp`, `openai`, `trl`, `sentence-transformers`.
            task: the NLP task to use for training. Currently supported tasks are: `TrainingTaskForTextClassification`,
                `TrainingTaskForSFT`, `TrainingTaskForRM`, `TrainingTaskForPPO`, `TrainingTaskForDPO`, `TrainingTaskForSentenceSimilarity`.
            train_size: the size of the train set. If `None`, the whole dataset will be used for training.
            test_size: the size of the test set. If `None`, the whole dataset will be used for testing.
            seed: the seed to use for splitting the dataset into train and test sets.
            lang: the spaCy language to use for training. If `None`, the language of the dataset will be used.
        """
        warnings.warn(
            (
                "A local `FeedbackDataset` returned because "
                "`prepare_for_training` is not supported for `RemoteFeedbackDataset`. "
                "`RemoteFeedbackDataset`.pull().prepare_for_training(*args, **kwargs)` is applied."
            ),
            UserWarning,
        )
        local = self.pull()
        return local.prepare_for_training(
            framework=framework,
            task=task,
            train_size=train_size,
            test_size=test_size,
            seed=seed,
            lang=lang,
        )

    def push_to_argilla(
        self, name: str, workspace: Optional[Union[str, "Workspace"]] = None, show_progress: bool = False
    ) -> "RemoteFeedbackDataset":
        warnings.warn(
            "Already pushed datasets cannot be pushed to Argilla again because they are synced automatically.",
            UserWarning,
        )
        return self

    # TODO(@frascuchon): is this actually needed? what are the probabilities on a missmatch happening?
    def push_to_huggingface(self, repo_id: str, generate_card: Optional[bool] = True, *args, **kwargs) -> None:
        """Pushes the current `FeedbackDataset` to HuggingFace Hub.

        Note:
            The records from the `RemoteFeedbackDataset` are being pulled before pushing,
            to ensure that there's no missmatch while uploading those as those are lazily fetched.

        Args:
            repo_id: the ID of the HuggingFace repo to push the dataset to.
            generate_card: whether to generate a dataset card or not. Defaults to `True`.
        """
        warnings.warn(
            (
                "The dataset is first pulled locally and pushed to Hugging Face after because "
                "`push_to_huggingface` is not supported for a `RemoteFeedbackDataset`. "
                "`RemoteFeedbackDataset.pull().push_to_huggingface(...)` is applied."
            ),
            UserWarning,
        )
        dataset = self.pull()
        dataset.push_to_huggingface(repo_id=repo_id, generate_card=generate_card, *args, **kwargs)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
import logging
import platform
import uuid
from typing import TYPE_CHECKING, Any, Dict, Optional

from argilla.pydantic_v1 import BaseSettings

_DEFAULT_TELEMETRY_KEY = "C6FkcaoCbt78rACAgvyBxGBcMB3dM3nn"


class TelemetrySettings(BaseSettings):
    """
    Telemetry settings

    This settings class is defined here to not depend on the server settings.
    """

    enable_telemetry: bool = True
    telemetry_key: str = _DEFAULT_TELEMETRY_KEY

    class Config:
        env_prefix = "ARGILLA_"


telemetry_settings = TelemetrySettings()


try:
    from analytics import Client  # This import works only for version 2.2.0
except (ImportError, ModuleNotFoundError):
    # TODO: show some warning info
    telemetry_settings.enable_telemetry = False
    Client = None


_LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class TelemetryClient:
    enable_telemetry: dataclasses.InitVar[bool] = telemetry_settings.enable_telemetry
    disable_send: dataclasses.InitVar[bool] = False
    api_key: dataclasses.InitVar[str] = telemetry_settings.telemetry_key
    host: dataclasses.InitVar[str] = "https://api.segment.io"

    _machine_id: Optional[uuid.UUID] = dataclasses.field(init=False, default=None)

    @property
    def machine_id(self) -> uuid.UUID:
        return self._machine_id

    def __post_init__(self, enable_telemetry: bool, disable_send: bool, api_key: str, host: str):
        from argilla import __version__

        self.client = None
        if enable_telemetry:
            try:
                self.client = Client(write_key=api_key, gzip=True, host=host, send=not disable_send, max_retries=10)
            except Exception as err:
                _LOGGER.warning(f"Cannot initialize telemetry. Error: {err}. Disabling...")

        self._machine_id = uuid.UUID(int=uuid.getnode())
        self._system_info = {
            "system": platform.system(),
            "machine": platform.machine(),
            "platform": platform.platform(),
            "python_version": platform.python_version(),
            "sys_version": platform.version(),
            "version": __version__,
        }

    def track_data(self, action: str, data: Dict[str, Any], include_system_info: bool = True):
        if not self.client:
            return

        event_data = data.copy()
        self.client.track(
            user_id=str(self._machine_id),
            event=action,
            properties=event_data,
            context=self._system_info if include_system_info else {},
        )


_CLIENT = TelemetryClient()


def get_current_filename() -> Optional[str]:
    """Returns the filename of the current file.

    It will try to get the filename from the following sources:
    - __file__ variable (only works when running from python)
    - __vsc_ipynb_file__ variable (only works when running from vscode)
    - ipynbname.name() (only works when running from a notebook/gooble colab)
    - None if it can't be determined
    """
    from pathlib import Path

    try:
        try:
            # Should work if we are running from python
            return Path(__file__).stem
        except NameError as e:
            # This should work if we are running a notebook from vscode
            globals_ = globals()
            return Path(globals_["__vsc_ipynb_file__"]).stem
    except KeyError as e:
        # This should work for notebooks running locally or using google colab
        import urllib.parse

        import ipynbname

        return Path(urllib.parse.unquote_plus(ipynbname.name())).stem


def tutorial_running() -> None:
    """Can be called when a tutorial is executed so that the tutorial_id is used to identify the tutorial and send an event."""
    if tutorial_id := get_current_filename():
        _CLIENT.track_data(action="TutorialRunning", data={"tutorial_id": tutorial_id})


def get_telemetry_client() -> TelemetryClient:
    return _CLIENT



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

import typer

from argilla.cli.callback import init_callback

from .create import create_user
from .delete import delete_user
from .list import list_users

app = typer.Typer(help="Commands for user management", no_args_is_help=True, callback=init_callback)
_COMMANDS_REQUIRING_USER = ["delete"]


def callback(
    ctx: typer.Context,
    username: Optional[str] = typer.Option(None, help="Username of the user to which apply the command."),
) -> None:
    init_callback()

    if ctx.invoked_subcommand not in _COMMANDS_REQUIRING_USER:
        return

    if username is None:
        raise typer.BadParameter("The command requires a username provided using '--username' option")

    from argilla.cli.rich import echo_in_panel
    from argilla.client.users import User

    try:
        user = User.from_name(username)
    except ValueError as e:
        echo_in_panel(
            f"User with username={username} doesn't exist.", title="User not found", title_align="left", success=False
        )
        raise typer.Exit(code=1) from e
    except RuntimeError as e:
        echo_in_panel(
            "An unexpected error occurred when trying to get the user.",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    ctx.obj = user


app = typer.Typer(help="Holds CLI commands for user management.", no_args_is_help=True, callback=callback)

app.command(name="create", help="Creates a new user")(create_user)
app.command(name="delete", help="Deletes a user")(delete_user)
app.command(name="list", help="List users")(list_users)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Optional

from typing_extensions import Literal

from argilla.client.feedback.training.base import ArgillaTrainerSkeleton
from argilla.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla.training.spacy import ArgillaSpaCyTrainer as ArgillaSpaCyTrainerV1
from argilla.training.spacy import ArgillaSpaCyTransformersTrainer as ArgillaSpaCyTransformersTrainerV1
from argilla.training.spacy import _ArgillaSpaCyTrainerBase as _ArgillaSpaCyTrainerBaseV1
from argilla.utils.dependency import require_dependencies, requires_dependencies

if TYPE_CHECKING:
    from argilla.client.feedback.integrations.huggingface.model_card import (
        SpacyModelCardData,
        SpacyTransformersModelCardData,
    )


class _ArgillaSpaCyTrainerBase(_ArgillaSpaCyTrainerBaseV1, ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaSpaCyTrainer")
    _logger.setLevel(logging.INFO)

    require_dependencies("spacy")

    def __init__(
        self,
        language: Optional[str] = None,
        gpu_id: Optional[int] = -1,
        model: Optional[str] = None,
        optimize: Literal["efficiency", "accuracy"] = "efficiency",
        *args,
        **kwargs,
    ) -> None:
        """Initialize the `_ArgillaSpaCyTrainerBase` class.

        Args:
            dataset: A `spacy.tokens.DocBin` object or a tuple of `spacy.tokens.DocBin` objects.
            record_class:
                A `TextClassificationRecord`, `TokenClassificationRecord`, or `Text2TextRecord`
                object. Defaults to None.
            seed: A `int` with the seed for the random number generator. Defaults to None.
            multi_label: A `bool` indicating whether the task is multi-label or not. Defaults to False.
            language:
                A `str` with the `spaCy` language code e.g. "en". See all the supported languages and their
                codes in `spaCy` at https://spacy.io/usage/models#languages. Defaults to None.
            gpu_id:
                the GPU ID to use. Defaults to -1, which means that the CPU will be used by default.
                GPU IDs start in 0, which stands for the default GPU in the system, if available.
            model:
                A `str` with the `spaCy` model name to use. If it contains vectors it
                can also be used for training/fine-tuning, e.g. "en_core_web_lg"
                contains vectors, while "en_core_web_sm" doesn't. Defaults to None.
            optimize:
                A `str` with the optimization strategy to use. Either "efficiency" or "accuracy".
                Defaults to "efficiency", which means that the model will be smaller, faster,
                and use less memory, but it will be less accurate. If "accuracy" is used, the model
                will be larger, slower, and use more memory, but it will be more accurate.
                Defaults to "efficiency".

        Raises:
            NotImplementedError: If the `record_class` is not supported or if the
                `init_training_args` method has not been implemented.
        """
        ArgillaTrainerSkeleton.__init__(self, *args, **kwargs)
        import spacy

        self._model = model

        if self._record_class == TokenClassificationRecord:
            self._column_mapping = {
                "text": "text",
                "token": "tokens",
                "ner_tags": "ner_tags",
            }
            self._spacy_pipeline_components = ["ner"]
        elif self._record_class == TextClassificationRecord:
            if self._multi_label:
                self._column_mapping = {"text": "text", "binarized_label": "label"}
                self._spacy_pipeline_components = ["textcat_multilabel"]
            else:
                self._column_mapping = {"text": "text", "label": "label"}
                self._spacy_pipeline_components = ["textcat"]
        else:
            raise NotImplementedError("`rg.Text2TextRecord` is not supported yet.")

        self._train_dataset, self._eval_dataset = (
            self._dataset if isinstance(self._dataset, tuple) and len(self._dataset) > 1 else (self._dataset, None)
        )
        self._train_dataset_path = "./train.spacy"
        self._eval_dataset_path = "./dev.spacy" if self._eval_dataset else "./train.spacy"

        self.language = language or "en"
        self.optimize = optimize

        self.gpu_id = gpu_id
        self.use_gpu = False
        if self.gpu_id != -1:
            self.use_gpu = spacy.prefer_gpu(self.gpu_id)

        if self.use_gpu:
            try:
                require_dependencies("torch")
                self.has_torch = True
            except Exception:
                self.has_torch = False

            try:
                require_dependencies("tensorflow")
                self.has_tensorflow = True
            except Exception:
                self.has_tensorflow = False

            if not self.has_torch and not self.has_tensorflow:
                self._logger(
                    "Either `torch` or `tensorflow` need to be installed to use the"
                    " GPU, since any of those is required as the GPU allocator. Falling"
                    " back to the CPU."
                )
                self.use_gpu = False
                self.gpu_id = -1

        self.init_training_args()

    def get_trainer_kwargs(self):
        """Get the trainer kwargs to be used in the `spacy` trainer."""
        return self.trainer_kwargs["training"]

    @requires_dependencies("spacy-huggingface-hub")
    def push_to_huggingface(self, output_dir: str, **kwargs) -> str:
        r"""Uploads the model to [huggingface's model hub](https://huggingface.co/models).

        With spacy we don't need the `repo_id` as in the other frameworks, that
        variable is generated internally by `spacy_huggingface_hub`, we need the
        path to the nlp pipeline to package it and push it.

        See Also:
            The optional arguments are the following:
            namespace: Name of organization to which the pipeline should be uploaded.
            commit_msg: Commit message to use for update
            verbose: Output additional info for debugging, e.g. the full generated hub metadata.

        Args:
            output_dir: The same path passed to `save` method. The path where the nlp pipeline
                should be saved to.

        Returns:
            model_name:
                The model name will be used in the base trainer to find the repo to push the model card.
                If the url of a model is: https://huggingface.co/<NAMESPACE>/<MODEL_NAME>,
                pass <NAMESPACE>/<MODEL_NAME>.
        """
        from spacy.cli.package import package
        from spacy_huggingface_hub import push

        if self.trainer_model is None:
            raise ValueError(
                "No pipeline was initialized, you must call either `init_model` or `train` before calling this method."
            )

        output_dir = Path(output_dir)
        with TemporaryDirectory() as tmpdirname:
            output_dir_pkg = Path(tmpdirname) / "spacy-packaged"
            output_dir_pkg.mkdir(exist_ok=True, parents=True)

            if not output_dir.is_dir():
                raise ValueError(
                    f"output_dir: '{output_dir.resolve()}' doesn't exist, you must pass the path to the folder of the trained model."
                )
            self._logger.info("Packaging nlp pipeline")
            package(
                input_dir=output_dir.resolve(),
                output_dir=output_dir_pkg,
                create_sdist=False,
                create_wheel=True,
                name=output_dir.stem,
            )
            self._logger.info(f"spacy pipeline packaged at: {output_dir_pkg}")

            # The following line obtains the full path to the .whl file:
            # The output dir contains a single package name. Inside this package
            # there will be a `dist` folder containing the packages. As we always
            # force to generate only the `wheel` package option, there we can only
            # find the .whl file. In case we generated both the wheel and sdist,
            # we could find it by getting the file with .whl extension.
            whl_path = next((next(output_dir_pkg.iterdir()) / "dist").iterdir())
            # Remove unused parameters from push to avoid errors:
            expected_kwargs = set(("namespace", "commig_msg", "silent", "verbose"))
            for kw in tuple(kwargs.keys()):
                if kw not in expected_kwargs:
                    kwargs.pop(kw)

            self._logger.info(f"Pushing: {whl_path} to huggingface hub.")
            result = push(whl_path, **kwargs)
            url = result["url"]

        self._logger.info(f"Model pushed to: {url}")
        # Passing the model name generated with spacy-huggingface-hub to use
        # it in the base ArgillaTrainer, it's easier to grab
        # the generated repo name than forcing the user to pass an argument.
        return url.replace("https://huggingface.co/", "")


class ArgillaSpaCyTrainer(ArgillaSpaCyTrainerV1, _ArgillaSpaCyTrainerBase):
    def __init__(self, freeze_tok2vec: bool = False, **kwargs) -> None:
        """Initialize the `ArgillaSpaCyTrainer` class.

        Args:
            freeze_tok2vec: A `bool` indicating whether to freeze the `tok2vec` weights
                during the training. Defaults to False.
            **kwargs: The `ArgillaSpaCyTrainerBase` arguments.

        Examples:
            >>> from argilla import ArgillaSpaCyTrainer
            >>> trainer = ArgillaSpaCyTrainer(
        """
        self.freeze_tok2vec = freeze_tok2vec
        _ArgillaSpaCyTrainerBase.__init__(self, **kwargs)

    def get_model_card_data(self, **card_data_kwargs) -> "SpacyModelCardData":
        """
        Generate the card data to be used for the `ArgillaModelCard`.

        Args:
            card_data_kwargs: Extra arguments provided by the user when creating the `ArgillaTrainer`.

        Returns:
            SpacyModelCardData: Container for the data to be written on the `ArgillaModelCard`.
        """
        from argilla.client.feedback.integrations.huggingface.model_card import SpacyModelCardData

        return SpacyModelCardData(
            model_id=self._model,
            task=self._task,
            lang=self.language,
            gpu_id=self.gpu_id,
            framework_kwargs={"optimize": self.optimize, "freeze_tok2vec": self.freeze_tok2vec},
            pipeline=self._spacy_pipeline_components,  # Used only to keep track for the config arguments
            update_config_kwargs=self.trainer_kwargs["training"],
            **card_data_kwargs,
        )


class ArgillaSpaCyTransformersTrainer(ArgillaSpaCyTransformersTrainerV1, _ArgillaSpaCyTrainerBase):
    def __init__(self, update_transformer: bool = True, **kwargs) -> None:
        """Initialize the `ArgillaSpaCyTransformersTrainer` class.

        Args:
            update_transformer: A `bool` indicating whether to update the transformer
                weights during the training. Defaults to True.
            **kwargs: The `ArgillaSpaCyTrainerBase` arguments.
        """
        self.update_transformer = update_transformer
        _ArgillaSpaCyTrainerBase.__init__(self, **kwargs)

    def get_model_card_data(self, **card_data_kwargs) -> "SpacyTransformersModelCardData":
        """
        Generate the card data to be used for the `ArgillaModelCard`.

        Args:
            card_data_kwargs: Extra arguments provided by the user when creating the `ArgillaTrainer`.

        Returns:
            SpacyTransformersModelCardData: Container for the data to be written on the `ArgillaModelCard`.
        """
        from argilla.client.feedback.integrations.huggingface.model_card import SpacyTransformersModelCardData

        return SpacyTransformersModelCardData(
            model_id=self._model,
            task=self._task,
            lang=self.language,
            gpu_id=self.gpu_id,
            framework_kwargs={"optimize": self.optimize, "update_transformer": self.update_transformer},
            pipeline=self._spacy_pipeline_components,  # Used only to keep track for the config arguments
            update_config_kwargs=self.trainer_kwargs["training"],
            **card_data_kwargs,
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import sys
from functools import wraps
from typing import Any, Callable, Coroutine, Dict, Type, TypeVar

import typer

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec


P = ParamSpec("P")
R = TypeVar("R")

HandleErrorFunc = Callable[[Exception], None]


class ArgillaTyper(typer.Typer):
    error_handlers: Dict[Type[Exception], HandleErrorFunc] = {}

    # https://github.com/tiangolo/typer/issues/88#issuecomment-1613013597
    def command(
        self, *args: Any, **kwargs: Any
    ) -> Callable[[Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]]:
        super_command = super().command(*args, **kwargs)

        def decorator(func: Callable[P, Coroutine[Any, Any, R]]) -> Callable[P, Coroutine[Any, Any, R]]:
            @wraps(func)
            def sync_func(*_args: P.args, **_kwargs: P.kwargs) -> R:
                return asyncio.run(func(*_args, **_kwargs))

            if asyncio.iscoroutinefunction(func):
                super_command(sync_func)
            else:
                super_command(func)

            return func

        return decorator

    def error_handler(self, exc: Type[Exception]) -> Callable[[HandleErrorFunc], None]:
        def decorator(func: HandleErrorFunc) -> None:
            self.error_handlers[exc] = func

        return decorator

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        try:
            return super().__call__(*args, **kwargs)
        except typer.Exit as e:
            raise e
        except Exception as e:
            handler = self.error_handlers.get(type(e))
            if handler is None:
                raise e
            handler(e)


def run(function: Callable[..., Coroutine[Any, Any, Any]]) -> None:
    app = ArgillaTyper(add_completion=False)
    app.command()(function)
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

import typer

app = typer.Typer(invoke_without_command=True)


@app.callback(help="Login to an Argilla Server")
def login(
    api_url: str = typer.Option(..., help="The URL of the Argilla Server to login in to"),
    api_key: str = typer.Option(
        ..., prompt="API Key", hide_input=True, help="The API key for logging into the Argilla Server"
    ),
    workspace: Optional[str] = typer.Option(
        None, help="The default workspace over which the operations will be performed"
    ),
    extra_headers: Optional[str] = typer.Option(
        None, help="A JSON string with extra headers to be sent in the requests to the Argilla Server"
    ),
):
    import json

    from argilla.cli.rich import echo_in_panel
    from argilla.client.login import login as login_func

    try:
        if extra_headers:
            headers = json.loads(extra_headers)
        else:
            headers = {}
        login_func(api_url=api_url, api_key=api_key, workspace=workspace, extra_headers=headers)
    except json.JSONDecodeError as e:
        echo_in_panel(
            "The provided extra headers are not a valid JSON string.",
            title="Extra headers error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e
    except ValueError as e:
        echo_in_panel(
            f"Could not login to the '{api_url}' Argilla server. Please check the provided credentials and try again.",
            title="Login error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    echo_in_panel(f"Logged in successfully to '{api_url}' Argilla server!", title="Logged in", title_align="left")


if __name__ == "__main__":
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import re
import warnings
from typing import List, Optional

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

import json

from argilla.pydantic_v1 import BaseModel, Field

try:
    from yaml import SafeLoader, load, safe_dump

except ImportError:
    raise ImportError(
        "Please make sure to install `PyYAML` in order to use `DatasetConfig`. To do"
        " so you can run `pip install pyyaml`."
    )

from argilla.client.feedback.schemas.types import AllowedFieldTypes, AllowedMetadataPropertyTypes, AllowedQuestionTypes
from argilla.client.feedback.schemas.vector_settings import VectorSettings


class DatasetConfig(BaseModel):
    fields: List[AllowedFieldTypes]
    questions: List[Annotated[AllowedQuestionTypes, Field(..., discriminator="type")]]
    guidelines: Optional[str] = None
    metadata_properties: Optional[
        List[Annotated[AllowedMetadataPropertyTypes, Field(..., discriminator="type")]]
    ] = None
    allow_extra_metadata: bool = True
    vectors_settings: Optional[List[VectorSettings]] = None

    class Config:
        use_enum_values = True

    def to_yaml(self) -> str:
        # THIS DOUBLE SERIALIZATION IS DONE TO AVOID GENERATING YAMLs WITH CUSTOM ENUM TYPES.
        # SEE https://github.com/argilla-io/argilla/issues/4089
        # A better solution must be applied. We must review all defined enums in Pydantic models and set up the
        # `use_enum_values = True` flag.
        # See also https://github.com/yaml/pyyaml/issues/722
        return safe_dump(json.loads(self.json()))

    @classmethod
    def from_yaml(cls, yaml_str: str) -> "DatasetConfig":
        yaml_str = re.sub(r"(\n\s*|)id: !!python/object:uuid\.UUID\s+int: \d+", "", yaml_str)
        yaml_dict = load(yaml_str, Loader=SafeLoader)
        # Here for backwards compatibility
        for field in yaml_dict["fields"]:
            field.pop("id", None)
            field.pop("settings", None)
        for question in yaml_dict["questions"]:
            question.pop("id", None)
            question.pop("settings", None)
        return cls(**yaml_dict)


# TODO(alvarobartt): here for backwards compatibility, last used in v1.12.0
class DeprecatedDatasetConfig(BaseModel):
    fields: List[AllowedFieldTypes]
    questions: List[AllowedQuestionTypes]
    guidelines: Optional[str] = None

    def to_json(self) -> str:
        warnings.warn(
            "`DatasetConfig` can just be dumped to YAML, so make sure that you are"
            " dumping to a YAML file instead of a JSON file. `DatasetConfig` will come"
            " in YAML format from now on, instead of JSON format.",
            DeprecationWarning,
        )
        return self.json()

    @classmethod
    def from_json(cls, json_str: str) -> "DeprecatedDatasetConfig":
        warnings.warn(
            "`DatasetConfig` can just be loaded from YAML, so make sure that you are"
            " loading a YAML file instead of a JSON file. `DatasetConfig` will be dumped"
            " as YAML from now on, instead of JSON.",
            DeprecationWarning,
        )
        parsed_json = json.loads(json_str)
        # Here for backwards compatibility
        for field in parsed_json["fields"]:
            # for 1.10.0, 1.9.0, and 1.8.0
            field.pop("id", None)
            field.pop("inserted_at", None)
            field.pop("updated_at", None)
            if "settings" not in field:
                continue
            field["type"] = field["settings"]["type"]
            if "use_markdown" in field["settings"]:
                field["use_markdown"] = field["settings"]["use_markdown"]
            # for 1.12.0 and 1.11.0
            field.pop("settings", None)
        for question in parsed_json["questions"]:
            # for 1.10.0, 1.9.0, and 1.8.0
            question.pop("id", None)
            question.pop("inserted_at", None)
            question.pop("updated_at", None)
            if "settings" not in question:
                continue
            question.update({"type": question["settings"]["type"]})
            if question["type"] in ["rating", "ranking"]:
                question["values"] = [option["value"] for option in question["settings"]["options"]]
            elif question["type"] in ["label_selection", "multi_label_selection"]:
                if all(option["value"] == option["text"] for option in question["settings"]["options"]):
                    question["labels"] = [option["value"] for option in question["settings"]["options"]]
                else:
                    question["labels"] = {option["value"]: option["text"] for option in question["settings"]["options"]}
                if "visible_labels" in question["settings"]:
                    question["visible_labels"] = question["settings"]["visible_labels"]
            # for 1.12.0 and 1.11.0
            question.pop("settings", None)
        return cls(**parsed_json)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
import re
import warnings
from asyncio import Future
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import backoff
import httpx
from rich import print as rprint
from rich.progress import Progress

from argilla._constants import (
    DATASET_NAME_REGEX_PATTERN,
    DEFAULT_API_KEY,
    DEFAULT_API_URL,
    DEFAULT_USERNAME,
    WORKSPACE_HEADER_NAME,
    WORKSPACE_NAME_REGEX_PATTERN,
)
from argilla.client.apis.datasets import Datasets
from argilla.client.apis.metrics import MetricsAPI
from argilla.client.apis.search import Search, VectorSearch
from argilla.client.apis.status import Status
from argilla.client.datasets import (
    Dataset,
    DatasetForText2Text,
    DatasetForTextClassification,
    DatasetForTokenClassification,
)
from argilla.client.metrics.models import MetricResults
from argilla.client.models import (
    BulkResponse,
    Record,
    Text2TextRecord,
    TextClassificationRecord,
    TokenClassificationRecord,
)
from argilla.client.sdk.client import AuthenticatedClient
from argilla.client.sdk.commons.api import bulk
from argilla.client.sdk.commons.errors import AlreadyExistsApiError, InputValueError, NotFoundApiError
from argilla.client.sdk.datasets import api as datasets_api
from argilla.client.sdk.datasets.models import CopyDatasetRequest, TaskType
from argilla.client.sdk.datasets.models import Dataset as DatasetModel
from argilla.client.sdk.metrics import api as metrics_api
from argilla.client.sdk.metrics.models import MetricInfo
from argilla.client.sdk.text2text.models import (
    CreationText2TextRecord,
    Text2TextBulkData,
)
from argilla.client.sdk.text2text.models import (
    Text2TextRecord as SdkText2TextRecord,
)
from argilla.client.sdk.text_classification import api as text_classification_api
from argilla.client.sdk.text_classification.models import (
    CreationTextClassificationRecord,
    LabelingRule,
    LabelingRuleMetricsSummary,
    TextClassificationBulkData,
)
from argilla.client.sdk.text_classification.models import (
    TextClassificationRecord as SdkTextClassificationRecord,
)
from argilla.client.sdk.token_classification.models import (
    CreationTokenClassificationRecord,
    TokenClassificationBulkData,
)
from argilla.client.sdk.token_classification.models import (
    TokenClassificationRecord as SdkTokenClassificationRecord,
)
from argilla.client.sdk.users import api as users_api
from argilla.client.sdk.v1.workspaces import api as workspaces_api_v1
from argilla.client.sdk.v1.workspaces.models import WorkspaceModel

_LOGGER = logging.getLogger(__name__)


class Argilla:
    """
    The main argilla client.
    """

    # Larger sizes will trigger a warning
    _MAX_BATCH_SIZE = 5000

    def __init__(
        self,
        api_url: Optional[str] = None,
        api_key: Optional[str] = None,
        workspace: Optional[str] = None,
        timeout: int = 120,
        extra_headers: Optional[Dict[str, str]] = None,
        httpx_extra_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """
        Inits `Argilla` instance.

        If called with `api_url=None` and `api_key=None` and no values have been set for the environment variables
        `ARGILLA_API_URL` and `ARGILLA_API_KEY`, then the local credentials stored by a previous call to `argilla login`
        command will be used. If local credentials are not found, then `api_url` and `api_key` will fallback to the
        default values.

        Args:
            api_url: Address of the REST API. If `None` (default) and the env variable ``ARGILLA_API_URL`` is not set,
                it will default to `http://localhost:6900`.
            api_key: Authentication key for the REST API. If `None` (default) and the env variable ``ARGILLA_API_KEY``
                is not set, it will default to `argilla.apikey`.
            workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
                env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
            timeout: Wait `timeout` seconds for the connection to timeout. Default: 60.
            extra_headers: Extra HTTP headers sent to the server. You can use this to customize
                the headers of argilla client requests, like additional security restrictions. Default: `None`.
            httpx_extra_kwargs: Extra kwargs passed to the `httpx.Client` constructor. For more information about the
                available arguments, see https://www.python-httpx.org/api/#client. Defaults to `None`.
        """
        from argilla.client.login import ArgillaCredentials

        api_url = api_url or os.getenv("ARGILLA_API_URL")
        api_key = api_key or os.getenv("ARGILLA_API_KEY")
        workspace = workspace or os.getenv("ARGILLA_WORKSPACE")
        extra_headers = extra_headers or {}

        if api_url is None and api_key is None:
            try:
                credentials = ArgillaCredentials.load()
                api_url = credentials.api_url
                api_key = credentials.api_key
                if not workspace:
                    workspace = credentials.workspace
                extra_headers = credentials.extra_headers
            except FileNotFoundError:
                pass

        api_url = api_url or DEFAULT_API_URL
        api_key = api_key or DEFAULT_API_KEY

        # Checking that the api_url does not end in '/'
        api_url = re.sub(r"\/$", "", api_url)
        headers = extra_headers or {}

        self._client: AuthenticatedClient = AuthenticatedClient(
            base_url=api_url,
            token=api_key,
            timeout=timeout,
            headers=headers.copy(),
            httpx_extra_kwargs=httpx_extra_kwargs,
        )

        self._user = users_api.whoami(client=self.http_client)  # .parsed

        if not workspace and self._user.username == DEFAULT_USERNAME and DEFAULT_USERNAME in self._user.workspaces:
            warnings.warn(
                "Default user was detected and no workspace configuration was provided,"
                f" so the default {DEFAULT_USERNAME!r} workspace will be used. If you"
                " want to setup another workspace, use the `rg.set_workspace` function"
                " or provide a different one on `rg.init`",
                category=UserWarning,
            )
            workspace = DEFAULT_USERNAME
        if workspace:
            self.set_workspace(workspace or self._user.username)
        else:
            warnings.warn(
                "No workspace configuration was detected. To work with Argilla"
                " datasets, specify a valid workspace name on `rg.init` or set it"
                " up through the `rg.set_workspace` function.",
                category=UserWarning,
            )

        self._check_argilla_versions()

    def _check_argilla_versions(self):
        from argilla import __version__ as rg_version

        api_info = Status(self.http_client).get_info()

        client_version = rg_version.split(".")[:2]
        server_version = api_info.version.split(".")[:2]
        if client_version != server_version:
            warnings.warn(
                message=f"You're connecting to Argilla Server {api_info.version} using a different client version "
                f"({rg_version}).\n"
                "This may lead to potential compatibility issues during your experience.\n"
                "To ensure a seamless and optimized connection, we highly recommend "
                "aligning your client version with the server version.",
                category=UserWarning,
            )

    def __del__(self):
        if hasattr(self, "_client"):
            del self._client

    @property
    def client(self) -> AuthenticatedClient:
        """The underlying authenticated HTTP client"""
        warnings.warn(
            message="This prop will be removed in next release. Please use the http_client prop instead.",
            category=UserWarning,
        )
        return self._client

    @property
    def http_client(self):
        """The underlying authenticated HTTP client"""
        return self._client

    @property
    def datasets(self) -> Datasets:
        """Datasets API primitives"""
        return Datasets(client=self.http_client)

    @property
    def search(self):
        """Search API primitives"""
        return Search(client=self.http_client)

    @property
    def metrics(self):
        """Metrics API primitives"""
        return MetricsAPI(client=self.http_client)

    @property
    def user(self):
        """The current connected user"""
        return self._user

    def set_workspace(self, workspace: str):
        """Sets the active workspace.

        Args:
            workspace: The new workspace
        """
        if not workspace:
            raise Exception("Must provide a workspace")

        if not re.match(WORKSPACE_NAME_REGEX_PATTERN, workspace):
            raise InputValueError(
                f"Provided workspace name {workspace} does not match the pattern"
                f" {WORKSPACE_NAME_REGEX_PATTERN}. Please, use a valid name for your"
                " workspace. This limitation is caused by naming conventions for indexes"
                " in Elasticsearch. If applicable, you can try to lowercase the name of your workspace."
                " https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-create-index.html"
            )

        if workspace != self.get_workspace():
            if workspace in [ws.name for ws in self.list_workspaces()]:
                self.http_client.update_headers({WORKSPACE_HEADER_NAME: workspace})
            else:
                raise ValueError(f"Wrong provided workspace {workspace!r}")

    def get_workspace(self) -> Optional[str]:
        """Returns the name of the active workspace.

        Returns:
            The name of the active workspace as a string.
        """
        return self.http_client.headers.get(WORKSPACE_HEADER_NAME)

    def list_workspaces(self) -> List[WorkspaceModel]:
        """Lists all the available workspaces for the current user.

        Returns:
            A list of `WorkspaceModel` objects, containing the workspace
            attributes: name, id, created_at, and updated_at.
        """
        return workspaces_api_v1.list_workspaces_me(client=self.http_client.httpx).parsed

    def list_datasets(self, workspace: Optional[str] = None) -> List[DatasetModel]:
        """Lists all the available datasets for the current user in Argilla.

        Args:
            workspace: If provided, list datasets from that workspace only. Note that
                the workspace must exist in advance, otherwise a HTTP 400 error will be
                raised.

        Returns:
            A list of `DatasetModel` objects, containing the dataset
            attributes: tags, metadata, name, id, task, owner, workspace, created_at,
            and last_updated.
        """
        return datasets_api.list_datasets(client=self.http_client, workspace=workspace).parsed

    def copy(self, dataset: str, name_of_copy: str, workspace: Optional[str] = None) -> None:
        """Creates a copy of a dataset including its tags and metadata

        Args:
            dataset: Name of the source dataset
            name_of_copy: Name of the copied dataset
            workspace: If provided, dataset will be copied to that workspace

        """
        datasets_api.copy_dataset(
            client=self.http_client,
            name=dataset,
            json_body=CopyDatasetRequest(name=name_of_copy, target_workspace=workspace),
        )

    def get_dataset(self, name: str, workspace: Optional[str] = None) -> DatasetModel:
        """Gets a dataset by name.

        Args:
            name: The dataset name.
            workspace: If provided, dataset will be retrieved from that workspace. Otherwise, the active workspace will
                be used.

        Returns:
            A `Dataset` object containing the dataset information.
        """
        response = datasets_api.get_dataset(client=self.http_client, name=name, workspace=workspace)
        return response.parsed

    def delete(self, name: str, workspace: Optional[str] = None):
        """Deletes a dataset.

        Args:
            name: The dataset name.
        """
        if workspace is not None:
            self.set_workspace(workspace)

        datasets_api.delete_dataset(client=self.http_client, name=name)

    def log(
        self,
        records: Union[Record, Iterable[Record], Dataset],
        name: str,
        workspace: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        batch_size: int = 100,
        verbose: bool = True,
        background: bool = False,
        num_threads: int = 0,
        max_retries: int = 3,
        chunk_size: Optional[int] = None,
    ) -> Union[BulkResponse, Future]:
        """Logs Records to argilla.

        The logging happens asynchronously in a background thread.

        Args:
            records: The record, an iterable of records, or a dataset to log.
            name: The dataset name.
            tags: A dictionary of tags related to the dataset.
            metadata: A dictionary of extra info for the dataset.
            batch_size: The batch size for a data bulk.
            verbose: If True, shows a progress bar and prints out a quick summary at the end.
            background: If True, we will NOT wait for the logging process to finish and return
                an ``asyncio.Future`` object. You probably want to set ``verbose`` to False
                in that case.
            num_threads: If > 0, will use num_thread separate number threads to batches, sending data concurrently.
                Default to `0`, which means no threading at all.
            max_retries: Number of retries when logging a batch of records if a `httpx.TransportError` occurs.
                Default `3`
            chunk_size: DEPRECATED! Use `batch_size` instead.

        Returns:
            Summary of the response from the REST API.
            If the ``background`` argument is set to True, an ``asyncio.Future``
            will be returned instead.

        """

        if background:
            executor = ThreadPoolExecutor(max_workers=1)

            return executor.submit(
                self.log,
                records=records,
                name=name,
                workspace=workspace,
                tags=tags,
                metadata=metadata,
                batch_size=batch_size,
                verbose=verbose,
                chunk_size=chunk_size,
                num_threads=num_threads,
                max_retries=max_retries,
            )

        tags = tags or {}
        metadata = metadata or {}

        if workspace is not None:
            self.set_workspace(workspace)

        if not name:
            raise InputValueError("Empty dataset name has been passed as argument.")

        if not re.match(DATASET_NAME_REGEX_PATTERN, name):
            raise InputValueError(
                f"Provided dataset name {name} does not match the pattern"
                f" {DATASET_NAME_REGEX_PATTERN}. Please, use a valid name for your"
                " dataset. This limitation is caused by naming conventions for indexes"
                " in Elasticsearch."
                " https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-create-index.html"
            )

        if chunk_size is not None:
            warnings.warn(
                "The argument `chunk_size` is deprecated and will be removed in a future"
                " version. Please use `batch_size` instead.",
                FutureWarning,
            )
            batch_size = chunk_size

        if batch_size > self._MAX_BATCH_SIZE:
            warnings.warn(
                "The requested batch size is noticeably large, timeout errors may occur. "
                f"Consider a batch size smaller than {self._MAX_BATCH_SIZE}",
            )

        if isinstance(records, Record.__args__):
            records = [records]
        records = list(records)

        try:
            record_type = type(records[0])
        except IndexError:
            raise InputValueError("Empty record list has been passed as argument.")

        if record_type is TextClassificationRecord:
            bulk_class = TextClassificationBulkData
            creation_class = CreationTextClassificationRecord
            task = TaskType.text_classification
        elif record_type is TokenClassificationRecord:
            bulk_class = TokenClassificationBulkData
            creation_class = CreationTokenClassificationRecord
            task = TaskType.token_classification
        elif record_type is Text2TextRecord:
            bulk_class = Text2TextBulkData
            creation_class = CreationText2TextRecord
            task = TaskType.text2text
        else:
            raise InputValueError(f"Unknown record type {record_type}. Available values are {Record.__args__}")

        try:
            self.datasets.create(name=name, task=task, workspace=workspace)
        except AlreadyExistsApiError:
            pass

        results = []
        with Progress() as progress_bar:
            task = progress_bar.add_task("Logging...", total=len(records), visible=verbose)

            batches = [records[i : i + batch_size] for i in range(0, len(records), batch_size)]

            @backoff.on_exception(
                backoff.expo,
                exception=httpx.TransportError,
                max_tries=max_retries,
                backoff_log_level=logging.DEBUG,
            )
            def log_batch(batch_info: Tuple[int, list]) -> Tuple[int, int]:
                batch_id, batch = batch_info

                bulk_result = bulk(
                    client=self.http_client,
                    name=name,
                    json_body=bulk_class(
                        tags=tags, metadata=metadata, records=[creation_class.from_client(r) for r in batch]
                    ),
                )

                progress_bar.update(task, advance=len(batch))
                return bulk_result.processed, bulk_result.failed

            if num_threads >= 1:
                with ThreadPoolExecutor(max_workers=num_threads) as executor:
                    results.extend(list(executor.map(log_batch, enumerate(batches))))
            else:
                results.extend(list(map(log_batch, enumerate(batches))))

        processed_batches, failed_batches = zip(*results)
        processed, failed = sum(processed_batches), sum(failed_batches)

        # TODO: improve logging policy in library
        if verbose:
            _LOGGER.info(f"Processed {processed} records in dataset {name}. Failed: {failed}")
            workspace = self.get_workspace()
            if not workspace:  # Just for backward comp. with datasets with no workspaces
                workspace = "-"
            url = f"{self.http_client.base_url}/datasets/{workspace}/{name}"
            rprint(f"{processed} records logged to [link={url}]{url}[/link]")

        # Creating a composite BulkResponse with the total processed and failed
        return BulkResponse(dataset=name, processed=processed, failed=failed)

    def delete_records(
        self,
        name: str,
        workspace: Optional[str] = None,
        query: Optional[str] = None,
        ids: Optional[List[Union[str, int]]] = None,
        discard_only: bool = False,
        discard_when_forbidden: bool = True,
    ) -> Tuple[int, int]:
        """Delete records from a argilla dataset.

        Args:
            name: The dataset name.
            query: An ElasticSearch query with the `query string syntax
                <https://docs.argilla.io/en/latest/practical_guides/filter_dataset.html>`_
            ids: If provided, deletes dataset records with given ids.
            discard_only: If `True`, matched records won't be deleted. Instead, they will be marked as `Discarded`
            discard_when_forbidden: Only super-user or dataset creator can delete records from a dataset.
                So, running "hard" deletion for other users will raise an `ForbiddenApiError` error.
                If this parameter is `True`, the client API will automatically try to mark as ``Discarded``
                records instead. Default, `True`

        Returns:
            The total of matched records and real number of processed errors. These numbers could not
            be the same if some data conflicts are found during operations (some matched records change during
            deletion).

        """
        if workspace is not None:
            self.set_workspace(workspace)

        return self.datasets.delete_records(
            name=name,
            mark_as_discarded=discard_only,
            discard_when_forbidden=discard_when_forbidden,
            query_text=query,
            ids=ids,
        )

    def load(
        self,
        name: str,
        workspace: Optional[str] = None,
        query: Optional[str] = None,
        vector: Optional[Tuple[str, List[float]]] = None,
        ids: Optional[List[Union[str, int]]] = None,
        limit: Optional[int] = None,
        sort: Optional[List[Tuple[str, str]]] = None,
        id_from: Optional[str] = None,
        batch_size: int = 250,
        include_vectors: bool = True,
        include_metrics: bool = True,
        as_pandas=None,
    ) -> Dataset:
        """Loads a argilla dataset.

        Args:
            name: The dataset name.
            query: An ElasticSearch query with the `query string
                syntax <https://docs.argilla.io/en/latesthttps://docs.argilla.io/en/latest/practical_guides/filter_dataset.html.html>`_
            vector: Vector configuration for a semantic search
            ids: If provided, load dataset records with given ids.
            limit: The number of records to retrieve.
            sort: The fields on which to sort [(<field_name>, 'asc|decs')].
            id_from: If provided, starts gathering the records starting from that Record.
                As the Records returned with the load method are sorted by ID, ´id_from´
                can be used to load using batches.
            batch_size: If provided, load `batch_size` samples per request. A lower batch
                size may help avoid timeouts.
            include_vectors: When set to `False`, indicates that the record data will be retrieved excluding its vectors,
                if any. By default, this parameter is set to `True`, meaning that vector data will be included.
            include_metrics: When set to `False`, indicates that the record data will be retrieved excluding its metrics.
                By default, this parameter is set to `True`, meaning that metrics will be included.
            as_pandas: DEPRECATED! To get a pandas DataFrame do
                ``rg.load('my_dataset').to_pandas()``.


        Returns:
            A argilla dataset.

        """
        if workspace is not None:
            self.set_workspace(workspace)

        if as_pandas is False:
            warnings.warn(
                "The argument `as_pandas` is deprecated and will be removed in a future"
                " version. Please adapt your code accordingly. ",
                FutureWarning,
            )
        elif as_pandas is True:
            raise ValueError(
                "The argument `as_pandas` is deprecated and will be removed in a future"
                " version. Please adapt your code accordingly. ",
                "If you want a pandas DataFrame do `rg.load('my_dataset').to_pandas()`.",
            )

        return self._load_records_internal(
            name=name,
            query=query,
            vector=vector,
            ids=ids,
            limit=limit,
            sort=sort,
            id_from=id_from,
            batch_size=batch_size,
            include_vectors=include_vectors,
            include_metrics=include_metrics,
        )

    def dataset_metrics(self, name: str) -> List[MetricInfo]:
        response = datasets_api.get_dataset(self.http_client, name)
        response = metrics_api.get_dataset_metrics(self.http_client, name=name, task=response.parsed.task)

        return response.parsed

    def get_metric(self, name: str, metric: str) -> Optional[MetricInfo]:
        metrics = self.dataset_metrics(name)
        for metric_ in metrics:
            if metric_.id == metric:
                return metric_

    def compute_metric(
        self,
        name: str,
        metric: str,
        query: Optional[str] = None,
        interval: Optional[float] = None,
        size: Optional[int] = None,
    ) -> MetricResults:
        response = datasets_api.get_dataset(self.http_client, name)

        metric_ = self.get_metric(name, metric=metric)
        assert metric_ is not None, f"Metric {metric} not found !!!"

        response = metrics_api.compute_metric(
            self.http_client,
            name=name,
            task=response.parsed.task,
            metric=metric,
            query=query,
            interval=interval,
            size=size,
        )

        return MetricResults(**metric_.dict(), results=response.parsed)

    def add_dataset_labeling_rules(self, dataset: str, rules: List[LabelingRule]):
        """Adds the dataset labeling rules"""
        for rule in rules:
            try:
                text_classification_api.add_dataset_labeling_rule(
                    self.http_client,
                    name=dataset,
                    rule=rule,
                )
            except AlreadyExistsApiError:
                _LOGGER.warning(f"Rule {rule} already exists. Please, update the rule instead.")
            except Exception as ex:
                _LOGGER.warning(f"Cannot create rule {rule}: {ex}")

    def update_dataset_labeling_rules(
        self,
        dataset: str,
        rules: List[LabelingRule],
    ):
        """Updates the dataset labeling rules"""
        for rule in rules:
            try:
                text_classification_api.update_dataset_labeling_rule(
                    self.http_client,
                    name=dataset,
                    rule=rule,
                )
            except NotFoundApiError:
                _LOGGER.info(f"Rule {rule} does not exists, creating...")
                text_classification_api.add_dataset_labeling_rule(self.http_client, name=dataset, rule=rule)
            except Exception as ex:
                _LOGGER.warning(f"Cannot update rule {rule}: {ex}")

    def delete_dataset_labeling_rules(self, dataset: str, rules: List[LabelingRule]):
        for rule in rules:
            try:
                text_classification_api.delete_dataset_labeling_rule(self.http_client, name=dataset, rule=rule)
            except Exception as ex:
                _LOGGER.warning(f"Cannot delete rule {rule}: {ex}")
        """Deletes the dataset labeling rules"""
        for rule in rules:
            text_classification_api.delete_dataset_labeling_rule(self.http_client, name=dataset, rule=rule)

    def fetch_dataset_labeling_rules(self, dataset: str) -> List[LabelingRule]:
        response = text_classification_api.fetch_dataset_labeling_rules(self.http_client, name=dataset)

        return [LabelingRule.parse_obj(data) for data in response.parsed]

    def rule_metrics_for_dataset(self, dataset: str, rule: LabelingRule) -> LabelingRuleMetricsSummary:
        response = text_classification_api.dataset_rule_metrics(
            self.http_client, name=dataset, query=rule.query, label=rule.label
        )

        return LabelingRuleMetricsSummary.parse_obj(response.parsed)

    def _load_records_internal(
        self,
        name: str,
        query: Optional[str] = None,
        vector: Optional[Tuple[str, List[float]]] = None,
        ids: Optional[List[Union[str, int]]] = None,
        limit: Optional[int] = None,
        sort: Optional[List[Tuple[str, str]]] = None,
        id_from: Optional[str] = None,
        batch_size: int = 250,
        include_vectors: bool = True,
        include_metrics: bool = True,
    ) -> Dataset:
        dataset = self.datasets.find_by_name(name=name)
        task = dataset.task

        task_config = {
            TaskType.text_classification: (SdkTextClassificationRecord, DatasetForTextClassification),
            TaskType.token_classification: (SdkTokenClassificationRecord, DatasetForTokenClassification),
            TaskType.text2text: (SdkText2TextRecord, DatasetForText2Text),
        }

        try:
            sdk_record_class, dataset_class = task_config[task]
        except KeyError:
            raise ValueError(
                f"Load method not supported for the '{task}' task. Supported Tasks: "
                f"{[TaskType.text_classification, TaskType.token_classification, TaskType.text2text]}"
            )

        if vector:
            if sort is not None:
                _LOGGER.warning("Results are sorted by vector similarity, so 'sort' parameter is ignored.")

            if not (include_metrics and include_vectors):
                _LOGGER.warning(
                    "Metrics and vectors cannot be excluded when using vector search. These parameters will be ignored."
                )

            vector_search = VectorSearch(name=vector[0], value=vector[1])
            results = self.search.search_records(
                name=name,
                task=task,
                size=limit or 100,
                # query args
                query_text=query,
                vector=vector_search,
            )

            return dataset_class(results.records)

        all_supported_fields = {
            "metadata.*",
            "status",
            "event_timestamp",
            "annotation*",
            "prediction*",
            "search_keywords",
            "inputs.*",
            "multi_label",
            "explanation*",
            "text",
            "tokens",
        }

        if include_vectors:
            all_supported_fields.add("vectors.*")

        if include_metrics:
            all_supported_fields.add("metrics.*")

        records = self.datasets.scan(
            name=name,
            projection=all_supported_fields,
            limit=limit,
            sort=sort,
            id_from=id_from,
            batch_size=batch_size,
            # Query
            query_text=query,
            ids=ids,
        )
        records = [sdk_record_class.parse_obj(r).to_client() for r in records]
        return dataset_class(records)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

from argilla import TokenClassificationRecord
from argilla.client.api import Api
from argilla.monitoring.base import BaseMonitor
from argilla.monitoring.types import MissingType

try:
    from flair import __version__ as _flair_version
    from flair.data import Sentence
    from flair.models import SequenceTagger
except (ImportError, ModuleNotFoundError):
    _flair_version = None
    Sentence = MissingType
    SequenceTagger = MissingType


class FlairMonitor(BaseMonitor):
    def _prepare_log_data(
        self,
        data: List[Tuple[Sentence, Dict[str, Any]]],
    ) -> Dict[str, Any]:
        return dict(
            records=[
                TokenClassificationRecord(
                    text=sentence.to_original_text(),
                    tokens=[token.text for token in sentence.tokens],
                    metadata=meta,
                    prediction_agent=self.agent,
                    event_timestamp=datetime.utcnow(),
                    prediction=[
                        (
                            label.value,
                            span.start_position,
                            span.end_position,
                            label.score,
                        )
                        for label, span in zip(
                            sentence.get_labels(self.__model__.tag_type), sentence.get_spans(self.__model__.tag_type)
                        )
                    ],
                )
                for sentence, meta in data
            ],
            name=self.dataset,
            tags={**(self.tags or {}), "flair_version": _flair_version},
        )

    def predict(self, sentences: Union[List[Sentence], Sentence], *args, **kwargs):
        metadata = kwargs.pop("metadata", None)
        result = self.__model__.predict(sentences, *args, **kwargs)

        if not isinstance(sentences, list):
            sentences = [sentences]

        if not metadata:
            metadata = [{}] * len(sentences)

        filtered_data = [(sentence, meta) for sentence, meta in zip(sentences, metadata) if self.is_record_accepted()]
        if filtered_data:
            self._log_future = self.send_records(filtered_data)

        return result


def flair_monitor(
    pl: SequenceTagger, api: Api, dataset: str, sample_rate: float, log_interval: float
) -> Optional[SequenceTagger]:
    return FlairMonitor(
        pl,
        api=api,
        dataset=dataset,
        sample_rate=sample_rate,
        log_interval=log_interval,
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional

import typer

from argilla.client.sdk.users.models import UserRole


def create_user(
    username: str = typer.Option(..., prompt=True, help="The username of the user to be created"),
    password: str = typer.Option(
        ..., prompt=True, confirmation_prompt=True, hide_input=True, help="The password of the user to be created"
    ),
    first_name: Optional[str] = typer.Option(None, help="The first name of the user to be created"),
    last_name: Optional[str] = typer.Option(None, help="The last name of the user to be created"),
    role: UserRole = typer.Option(UserRole.annotator, help="The role of the user to be created"),
    workspaces: Optional[List[str]] = typer.Option(
        None,
        "--workspace",
        help="A workspace name to which the user will be linked to. This option can be provided several times.",
    ),
) -> None:
    from rich.markdown import Markdown

    from argilla.cli.rich import echo_in_panel
    from argilla.client.users import User

    try:
        user = User.create(
            username=username,
            password=password,
            first_name=first_name,
            last_name=last_name,
            role=role,
            workspaces=workspaces,
        )
    except KeyError as e:
        echo_in_panel(
            f"User with name={username} already exists.",
            title="User already exists",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e
    except ValueError as e:
        echo_in_panel(
            f"Provided parameters are not valid:\n\n{e}", title="Invalid parameters", title_align="left", success=False
        )
        raise typer.Exit(code=1) from e
    except RuntimeError as e:
        echo_in_panel(
            "An unexpected error occurred when trying to create the user.",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    workspaces_str = ", ".join(workspace.name for workspace in user.workspaces)

    echo_in_panel(
        Markdown(
            f"- **Username**: {user.username}\n"
            f"- **Role**: {user.role}\n"
            f"- **First name**: {user.first_name}\n"
            f"- **Last name**: {user.last_name}\n"
            f"- **API Key**: {user.api_key}\n"
            f"- **Workspaces**: {workspaces_str}"
        ),
        title="User created",
        title_align="left",
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Dict, Optional, Type, Union

from argilla.client.feedback.schemas.enums import MetadataPropertyTypes
from argilla.client.feedback.schemas.metadata import (
    FloatMetadataProperty,
    IntegerMetadataProperty,
    TermsMetadataProperty,
)
from argilla.client.feedback.schemas.remote.shared import RemoteSchema

if TYPE_CHECKING:
    import httpx

    from argilla.client.sdk.v1.datasets.models import FeedbackMetadataPropertyModel


class RemoteTermsMetadataProperty(TermsMetadataProperty, RemoteSchema):
    def to_local(self) -> TermsMetadataProperty:
        return TermsMetadataProperty(
            name=self.name,
            title=self.title,
            visible_for_annotators=self.visible_for_annotators,
            values=self.values,
        )

    @classmethod
    def from_api(
        cls, payload: "FeedbackMetadataPropertyModel", client: Optional["httpx.Client"] = None
    ) -> "RemoteTermsMetadataProperty":
        return RemoteTermsMetadataProperty(
            client=client,
            id=payload.id,
            name=payload.name,
            title=payload.title,
            visible_for_annotators=payload.visible_for_annotators,
            values=payload.settings.get("values", None),
        )


class RemoteIntegerMetadataProperty(IntegerMetadataProperty, RemoteSchema):
    def to_local(self) -> IntegerMetadataProperty:
        return IntegerMetadataProperty(
            name=self.name,
            title=self.title,
            visible_for_annotators=self.visible_for_annotators,
            min=self.min,
            max=self.max,
        )

    @classmethod
    def from_api(
        cls, payload: "FeedbackMetadataPropertyModel", client: Optional["httpx.Client"] = None
    ) -> "RemoteIntegerMetadataProperty":
        return RemoteIntegerMetadataProperty(
            client=client,
            id=payload.id,
            name=payload.name,
            title=payload.title,
            visible_for_annotators=payload.visible_for_annotators,
            min=payload.settings.get("min", None),
            max=payload.settings.get("max", None),
        )


class RemoteFloatMetadataProperty(FloatMetadataProperty, RemoteSchema):
    def to_local(self) -> FloatMetadataProperty:
        return FloatMetadataProperty(
            name=self.name,
            title=self.title,
            visible_for_annotators=self.visible_for_annotators,
            min=self.min,
            max=self.max,
        )

    @classmethod
    def from_api(
        cls, payload: "FeedbackMetadataPropertyModel", client: Optional["httpx.Client"] = None
    ) -> "RemoteFloatMetadataProperty":
        return RemoteFloatMetadataProperty(
            client=client,
            id=payload.id,
            name=payload.name,
            title=payload.title,
            visible_for_annotators=payload.visible_for_annotators,
            min=payload.settings.get("min", None),
            max=payload.settings.get("max", None),
        )


RemoteMetadataPropertiesMapping: Dict[
    MetadataPropertyTypes,
    Type[Union[RemoteTermsMetadataProperty, RemoteIntegerMetadataProperty, RemoteFloatMetadataProperty]],
] = {
    MetadataPropertyTypes.terms: RemoteTermsMetadataProperty,
    MetadataPropertyTypes.integer: RemoteIntegerMetadataProperty,
    MetadataPropertyTypes.float: RemoteFloatMetadataProperty,
}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, List, Tuple, Union

from argilla.pydantic_v1 import BaseModel


class TextClassificationReturnTypes(BaseModel):
    """
    Union[
        Tuple[str, str],
        Tuple[str, List[str]],
        List[Tuple[str, str]],
        List[Tuple[str, List[str]]]
    ]
    """

    format: Union[
        Tuple[str, Union[str, int]],
        Tuple[str, List[Union[str, int]]],
        List[Tuple[str, Union[str, int]]],
        List[Tuple[str, List[Union[str, int]]]],
    ]


class SFTReturnTypes(BaseModel):
    """
    Union[
        str,
        List[str]
    ]
    """

    format: Union[str, List[str]]


class RMReturnTypes(BaseModel):
    """
    Union[
        Tuple[str, str],
        Tuple[str, List[str]],
        List[Tuple[str, str]],
        List[Tuple[str, List[str]]]
    ]
    """

    format: Union[Tuple[str, str], Tuple[str, List[str]], List[Tuple[str, str]], List[Tuple[str, List[str]]]]


class PPOReturnTypes(BaseModel):
    """
    Union[
        str,
        List[str]
    ]
    """

    format: Union[str, List[str]]


class DPOReturnTypes(BaseModel):
    """
    Union
        Tuple[str, str, str],
        List[Tuple[str, str, str]]
    ]
    """

    format: Union[Tuple[str, str, str], List[Tuple[str, str, str]]]


class QuestionAnsweringReturnTypes(BaseModel):
    """
    Union[
        Tuple[str, str, str],
        List[Tuple[str, str, str]]
    ]
    """

    format: Union[Tuple[str, str, str], List[Tuple[str, str, str]]]


class ChatCompletionReturnTypes(BaseModel):
    """
    Union[
        Tuple[str, str, str, str],
        List[Tuple[str, str, str, str]]
    ]
    """

    format: Union[Tuple[str, str, str, str], List[Tuple[str, str, str, str]]]


class SentenceSimilarityReturnTypes(BaseModel):
    r"""
    Union[
        Dict[str, Union[float, int, str]],  # case 1 with with two string elements and one int/float, case 3 with one or three strings and one int/float.
        List[Dict[str, Union[float, int, str]]],  # case 1 with with two string elements and one int/float, case 3 with one or three strings and one int/float.
    ]

    For a reference of the different cases take a look at:
    https://huggingface.co/blog/how-to-train-sentence-transformers#how-to-prepare-your-dataset-for-training-a-sentence-transformers-model
    """

    format: Union[Dict[str, Union[float, int, str]], List[Dict[str, Union[float, int, str]]]]



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import socket
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Generic, List, MutableMapping, Optional, TypeVar, Union
from uuid import uuid4

from argilla.client.models import Vectors as ClientVectors
from argilla.pydantic_v1 import BaseModel, Field, validator
from argilla.pydantic_v1.generics import GenericModel

if TYPE_CHECKING:
    from httpx import Response as HTTPXResponse

MACHINE_NAME = socket.gethostname()


class TaskStatus(str, Enum):
    default = "Default"
    edited = "Edited"
    discarded = "Discarded"
    validated = "Validated"


class BaseAnnotation(BaseModel):
    agent: str = Field(max_length=64)


T = TypeVar("T", bound=BaseAnnotation)


class VectorInfo(BaseModel):
    """Record vector info for api layer data model"""

    value: List[float]


SdkVectors = Dict[str, VectorInfo]


class SdkVectorSearch(BaseModel):
    """The sdk vector search parameters"""

    name: str
    value: List[float]


class BaseRecord(GenericModel, Generic[T]):
    id: Optional[Union[int, str]] = Field(default_factory=lambda: str(uuid4()))
    metadata: Dict[str, Any] = Field(default=None)
    event_timestamp: Optional[datetime] = None
    status: Optional[TaskStatus] = None
    prediction: Optional[T] = None
    annotation: Optional[T] = None
    vectors: Optional[SdkVectors] = None
    metrics: Dict[str, Any] = Field(default_factory=dict)
    search_keywords: Optional[List[str]] = None

    # this is a small hack to get a json-compatible serialization on cls.dict(), which we use for the httpx calls.
    # they want to build this feature into pydantic, see https://github.com/samuelcolvin/pydantic/issues/1409
    @validator("event_timestamp")
    def datetime_to_isoformat(cls, v: Optional[datetime]):
        if v is not None:
            return v.isoformat()

    @staticmethod
    def _from_client_vectors(vectors: ClientVectors) -> SdkVectors:
        sdk_vectors = None
        if vectors:
            sdk_vectors = {name: VectorInfo(value=vector) for name, vector in vectors.items()}
        return sdk_vectors

    @staticmethod
    def _to_client_vectors(vectors: SdkVectors) -> ClientVectors:
        client_vectors = None
        if vectors:
            client_vectors = {name: vector.value for name, vector in vectors.items()}
        return client_vectors


class UpdateDatasetRequest(BaseModel):
    tags: Dict[str, str] = Field(default_factory=dict)
    metadata: Dict[str, Any] = Field(default_factory=dict)


class PredictionStatus(str, Enum):
    OK = "ok"
    KO = "ko"


class ScoreRange(BaseModel):
    range_from: float = Field(default=0.0, alias="from")
    range_to: float = Field(default=None, alias="to")


R = TypeVar("R")


class Response(GenericModel, Generic[R]):
    status_code: int
    content: bytes
    headers: MutableMapping[str, str]
    parsed: Optional[R]

    @classmethod
    def from_httpx_response(cls, response: "HTTPXResponse") -> "Response[R]":
        return cls(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
        )


class BulkResponse(BaseModel):
    dataset: str
    processed: int
    failed: int = 0


class ErrorMessage(BaseModel):
    detail: str


class ValidationError(BaseModel):
    loc: List[str]
    msg: str
    type: str


class HTTPValidationError(BaseModel):
    detail: List[ValidationError]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer

from argilla.client.workspaces import Workspace


def create_workspace(
    name: str = typer.Argument(
        ...,
        help="The name of the workspace to be created",
    )
) -> None:
    """Creates a workspace for the logged user in Argilla"""
    from argilla.cli.rich import echo_in_panel

    try:
        Workspace.create(name=name)
    except ValueError as e:
        echo_in_panel(
            f"Workspace with name={name} already exists.",
            title="Workspace already exists",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e
    except RuntimeError as e:
        echo_in_panel(
            "An unexpected error occurred when trying to create the workspace.",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    echo_in_panel(
        f"Workspace with the name={name} successfully created.", title="Workspace created", title_align="left"
    )
